"""
Provide a simplified UI for hklpy diffractometer users.

The user must define a diffractometer instance, then
register that instance here calling `select_diffractometer(instance)`.

FUNCTIONS

.. autosummary::

    ~cahkl
    ~cahkl_table
    ~calc_UB
    ~list_samples
    ~new_sample
    ~or_swap
    ~select_diffractometer
    ~set_energy
    ~setor
    ~show_sample
    ~show_selected_diffractometer
    ~update_sample
    ~wh
    ~pa
"""

__all__ = """
    cahkl
    cahkl_table
    calc_UB
    change_sample
    list_samples
    new_sample
    or_swap
    pa
    select_diffractometer
    set_energy
    setor
    show_sample
    show_selected_diffractometer
    update_sample
    wh
""".split()

import logging
import gi

gi.require_version("Hkl", "5.0")
logger = logging.getLogger(__name__)

from .diffract import Diffractometer
from .util import Lattice
import numpy
import pyRestTable


_geom_ = None  # selected diffractometer geometry


def _check_geom_selected(*args, **kwargs):
    """Raise ValueError if no diffractometer geometry is selected."""
    if _geom_ is None:
        raise ValueError(
            "No diffractometer selected."
            " Call 'select_diffractometer(diffr)' where"
            " 'diffr' is a diffractometer instance."
        )


def cahkl(h, k, l):
    """
    Calculate motor positions for one reflection.

    Returns a namedtuple.
    Does not move motors.
    """
    _check_geom_selected()
    # TODO: make certain this will not move the motors!
    return _geom_.forward(h, k, l)


def cahkl_table(reflections, digits=5):
    """
    Print a table with motor positions for each reflection given.

    Parameters
    ----------
    reflections : list(tuple(number,number,number))
        This is a list of reflections where
        each reflection is a tuple of 3 numbers
        specifying (h, k, l) of the reflection
        to compute the ``forward()`` computation.

        Example:  ``[(1,0,0), (1,1,1)]``
    digits : int
        Number of digits to roundoff each position
        value.  Default is 5.
    """
    _check_geom_selected()
    print(_geom_.forward_solutions_table(reflections, digits=digits))


# def calc_energy():
#     # TODO: should this be added?
#     raise NotImplementedError


def calc_UB(r1, r2, wavelength=None):
    """Compute the UB matrix with two reflections."""
    _check_geom_selected()
    _geom_.calc.sample.compute_UB(r1, r2)
    print(_geom_.calc.sample.UB)


def change_sample(sample):
    """Pick a known sample to be the current selection."""
    _check_geom_selected()
    if sample not in _geom_.calc._samples:
        # fmt: off
        raise KeyError(
            f"Sample '{sample}' is unknown."
            f"  Known samples: {list(_geom_.calc._samples.keys())}"
        )
        # fmt: on
    _geom_.calc.sample = sample
    show_sample(sample)


def list_samples(verbose=True):
    """List all defined crystal samples."""
    _check_geom_selected()
    # always show the default sample first
    current_name = _geom_.calc.sample_name
    show_sample(current_name, verbose=verbose)

    # now, show any other samples
    for sample in _geom_.calc._samples.keys():
        if sample != current_name:
            if verbose:
                print("")
            show_sample(sample, verbose=verbose)


def new_sample(nm, a, b, c, alpha, beta, gamma):
    """Define a new crystal sample."""
    _check_geom_selected()
    if nm in _geom_.calc._samples:
        logger.warning(
            (
                "Sample '%s' is already defined."
                "  Use 'update_sample()' to change lattice parameters"
                " on the *current* sample."
            ),
            nm,
        )
    else:
        lattice = Lattice(a=a, b=b, c=c, alpha=alpha, beta=beta, gamma=gamma)
        _geom_.calc.new_sample(nm, lattice=lattice)
    show_sample()


def or_swap():
    """Swap the 2 [UB] reflections, re-compute & return new [UB]."""
    # note: or_swap is how the community of SPEC users knows this function
    _check_geom_selected()
    return _geom_.calc.sample.swap_orientation_reflections()


def select_diffractometer(instrument=None):
    """Name the diffractometer to be used."""
    global _geom_
    if instrument is None or isinstance(instrument, Diffractometer):
        _geom_ = instrument
    else:
        raise TypeError(f"{instrument} must be a 'Diffractometer' subclass")


def set_energy(value, units=None, offset=None):
    """
    Set the energy (thus wavelength) to be used.
    """
    _check_geom_selected()
    if units is not None:
        _geom_.energy_units.put(units)
    if offset is not None:
        _geom_.energy_offset.put(offset)
    _geom_.energy.put(value)


def setor(h, k, l, *args, wavelength=None, **kwargs):
    """Define a crystal reflection and its motor positions."""
    _check_geom_selected()
    if len(args) == 0:
        if len(kwargs) == 0:
            pos = _geom_.real_position
        else:
            # fmt: off
            pos = [
                kwargs[m]
                for m in _geom_.calc.physical_axis_names
                if m in kwargs
            ]
            # fmt: on
    else:
        pos = args

    # NOTE: libhkl gets the wavelength on a reflection from hkl.calc
    # when the wavelength is set, it calls libhkl directly
    # as self._geometry.wavelength_set(wavelength, self._units)
    # The code here uses that procedure.
    if wavelength not in (None, 0):
        _geom_.calc.wavelength = wavelength

    refl = _geom_.calc.sample.add_reflection(h, k, l, position=pos)
    return refl


def show_sample(sample_name=None, verbose=True):
    """Print the default sample name and crystal lattice."""
    _check_geom_selected()
    sample_name = sample_name or _geom_.calc.sample_name
    sample = _geom_.calc._samples[sample_name]

    title = sample_name
    if sample_name == _geom_.calc.sample.name:
        title += " (*)"

    # Print Lattice more simply (than as a namedtuple).
    lattice = [getattr(sample.lattice, parm) for parm in sample.lattice._fields]
    if verbose:
        tbl = pyRestTable.Table()
        tbl.addLabel("key")
        tbl.addLabel("value")
        tbl.addRow(("name", sample_name))
        tbl.addRow(("lattice", lattice))
        for i, r in enumerate(sample.reflections):
            tbl.addRow((f"reflection {i+1}", r))
        tbl.addRow(("U", numpy.round(sample.U, 5)))
        tbl.addRow(("UB", numpy.round(sample.UB, 5)))

        print(f"Sample: {title}\n")
        print(tbl)
    else:
        print(f"{title}: {lattice}")


def show_selected_diffractometer(instrument=None):
    """Print the name of the selected diffractometer."""
    if _geom_ is None:
        print("No diffractometer selected.")
    print(_geom_.name)


def update_sample(a, b, c, alpha, beta, gamma):
    """Update current sample lattice."""
    _check_geom_selected()
    _geom_.calc.sample.lattice = (
        a,
        b,
        c,
        alpha,
        beta,
        gamma,
    )  # define the current sample
    show_sample(_geom_.calc.sample.name, verbose=False)


def pa():
    """Report (all) the diffractometer settings."""
    _check_geom_selected()
    _geom_.pa()


def wh():
    """Report (brief) where is the diffractometer."""
    _check_geom_selected()
    _geom_.wh()



from bluesky import plans as bp
from bluesky.simulators import check_limits
from ophyd.positioner import LimitError
import gi
import numpy as np
import numpy.testing
import pytest


gi.require_version("Hkl", "5.0")
# NOTE: MUST call gi.require_version() BEFORE import hkl
from hkl import SimulatedE4CV


class Fourc(SimulatedE4CV):
    ...


@pytest.fixture(scope="function")
def fourc():
    fourc = Fourc("", name="fourc")
    return fourc


def test_initial(fourc):
    assert fourc.h.position == 0
    assert fourc.k.position == 0
    assert fourc.l.position == 0


def test_limits(fourc):
    assert fourc.h.limits == (0, 0)
    assert fourc.h.check_value(1) is None
    assert fourc.h.check_value(1.2) is None
    with pytest.raises(ValueError) as exinfo:
        fourc.h.check_value(10)
    assert "Unable to solve." in str(exinfo.value)


def test_check_value(fourc):
    assert fourc.check_value((1, 0, 0)) is None
    assert fourc.check_value(dict(h=1, k=0, l=0)) is None
    assert fourc.check_value(dict(h=1, k=0)) is None

    with pytest.raises(TypeError) as exinfo:
        assert fourc.check_value(1, 0, 0) is None
    assert "check_value() takes 2 positional arguments" in str(exinfo.value)

    with pytest.raises(ValueError) as exinfo:
        assert fourc.check_value(1) is None
    assert "Not all required values for a PseudoPosition" in str(exinfo.value)

    assert fourc.check_value(dict(h=1, tth=0)) is None

    with pytest.raises(LimitError) as exinfo:
        fourc.check_value(dict(h=1, tth=np.inf))
    assert "position=inf not within limits" in str(exinfo.value)

    with pytest.raises(KeyError) as exinfo:
        fourc.check_value(dict(waldo=0))
    assert "waldo not in fourc" in str(exinfo.value)


def test_move(fourc):
    ppos = (1.2, 1.2, 0.001)
    rpos = (
        58.051956519652485,
        44.99999005281953,
        89.95225352812487,
        116.10391303930497,
    )
    fourc.move(ppos)
    numpy.testing.assert_almost_equal(fourc.position, ppos)
    numpy.testing.assert_almost_equal(tuple(fourc.real_position), rpos)


def test_hl_scan(fourc):
    fourc.move((1.2, 1.2, 0.001))
    assert check_limits(bp.scan([fourc], fourc.h, 0.9, 1.1, fourc.l, 0, 0, 11)) is None


def test_h00_scan(fourc):
    fourc.move(1, 0, 0)
    assert check_limits(bp.scan([fourc], fourc.h, 0.9, 1.1, fourc.l, 0, 0, 11)) is None


def test_hkl_scan(fourc):
    fourc.move(1, 1, 1)
    assert (
        check_limits(
            # fmt: off
            bp.scan(
                [fourc],
                fourc.h, 0.9, 1.1,
                fourc.k, 0.9, 1.1,
                fourc.l, 0.9, 1.1,
                33,
            )
            # fmt: on
        )
        is None
    )


def test_hkl_range_error(fourc):
    with pytest.raises(ValueError) as exinfo:
        assert (
            check_limits(
                # fmt: off
                bp.scan(
                    [fourc],
                    fourc.h, 0.9, 1.1,
                    fourc.k, 0.9, 1.1,
                    fourc.l, 0.09, 123.1,
                    33,
                )
                # fmt: on
            )
            is None
        )
    assert "Unable to solve." in str(exinfo.value)


def test_real_axis(fourc):
    assert check_limits(bp.scan([fourc], fourc.tth, 10, 20, 3)) is None


def test_axis_contention(fourc):
    # contention if move pseudo and real positioners together
    with pytest.raises(ValueError) as exinfo:
        check_limits(bp.scan([fourc], fourc.tth, 10, 20, fourc.k, 0, 0, 3))
    assert "mix of real and pseudo" in str(exinfo.value)


def test_real_axis_range_multi(fourc):
    assert check_limits(bp.scan([fourc], fourc.tth, 10, 20, fourc.chi, 5, 7, 3)) is None


def test_real_axis_range_error(fourc):
    with pytest.raises(LimitError) as exinfo:
        check_limits(bp.scan([fourc], fourc.tth, 10, 20000, 3))
    assert "not within limits" in str(exinfo.value)



# NOTE: MUST call gi.require_version() BEFORE import hkl
import gi

gi.require_version("Hkl", "5.0")


from bluesky import RunEngine
from hkl.calc import A_KEV
from hkl import SimulatedE4CV
from hkl import SimulatedK4CV
from hkl import SI_LATTICE_PARAMETER
from ophyd.sim import hw
import bluesky.plans as bp
import databroker
import hkl.util
import numpy.testing
import pytest


class Fourc(SimulatedE4CV):
    ...


class Kappa(SimulatedK4CV):
    ...


@pytest.fixture
def cat():
    return databroker.temp().v2


@pytest.fixture
def RE(cat):
    engine = RunEngine({})
    engine.subscribe(cat.v1.insert)
    return engine


@pytest.fixture(scope="function")
def fourc():
    fourc = Fourc("", name="fourc")
    fourc.wait_for_connection()
    fourc._update_calc_energy()

    fourc.energy.put(A_KEV / 1.54)
    a0 = SI_LATTICE_PARAMETER
    sample = fourc.calc.new_sample("Si", lattice=(a0, a0, a0, 90, 90, 90))
    r_400 = sample.add_reflection(4, 0, 0, (-145.451, 0, 0, 69.0966))
    r_040 = sample.add_reflection(0, 4, 0, (-145.451, 0, 90, 69.0966))
    fourc.calc.sample.compute_UB(r_400, r_040)

    return fourc


@pytest.fixture(scope="function")
def kappa():
    kappa = Kappa("", name="kappa")
    kappa.wait_for_connection()
    kappa._update_calc_energy()

    kappa.energy.put(A_KEV / 1.54)
    a0 = SI_LATTICE_PARAMETER
    sample = kappa.calc.new_sample("Si", lattice=(a0, a0, a0, 90, 90, 90))
    r_400 = sample.add_reflection(4, 0, 0, (55.4507, 0, 90, -69.0966))
    r_040 = sample.add_reflection(0, 4, 0, (-1.5950, 134.7568, 123.3554, -69.0966))

    kappa.calc.sample.compute_UB(r_400, r_040)

    return kappa


def test_basic_setup(fourc, kappa):
    assert fourc is not None
    assert kappa is not None
    assert fourc != kappa
    assert fourc._reals != kappa._reals

    assert fourc.calc.wavelength == 1.54
    assert fourc.calc.wavelength == kappa.calc.wavelength


def test_fourc_orientation_save(cat, RE, fourc):
    assert len(cat) == 0
    det = hw().noisy_det

    # this run will not save orientation information
    _uids = RE(bp.count([det]))
    assert len(_uids) == 1
    assert len(cat) == 1
    assert "fourc" not in cat[1].primary.config

    # this run _will_ save orientation information
    _uids = RE(bp.count([det, fourc]))
    assert len(_uids) == 1
    assert len(cat) == 2
    xarray_data = cat[2].primary.config["fourc"].read()
    assert not isinstance(xarray_data, dict)
    descriptors = xarray_data.to_dict()
    assert isinstance(descriptors, dict)
    assert list(descriptors.keys()) == "coords attrs dims data_vars".split()
    key_list = """
        _pseudos
        _reals
        class_name
        diffractometer_name
        geometry_name
        lattice
        orientation_attrs
        reflections_details
        sample_name
        UB
    """.split()
    for key in key_list:
        key_name = f"fourc_{key}"
        assert hasattr(xarray_data, key_name)
        assert key_name in descriptors["data_vars"]

    assert xarray_data.fourc_class_name == "Fourc"
    assert xarray_data.fourc_geometry_name == "E4CV"
    assert xarray_data.fourc_diffractometer_name == "fourc"
    assert xarray_data.fourc_sample_name == "Si"

    assert len(xarray_data.fourc__pseudos) == 1
    assert xarray_data.fourc__pseudos[0].values.tolist() == "h k l".split()
    assert len(xarray_data.fourc__reals) == 1
    assert xarray_data.fourc__reals.values[0].tolist() == "omega chi phi tth".split()


def test_fourc_run_orientation_info(cat, RE, fourc):
    RE(bp.count([fourc]))
    info = hkl.util.run_orientation_info(cat[1])
    assert info is not None
    assert isinstance(info, dict)
    assert "fourc" in info
    fourc_orient = info["fourc"]
    assert "orientation_attrs" in fourc_orient
    assert "reflections_details" in fourc_orient["orientation_attrs"]
    assert "reflections_details" in fourc_orient
    refls = fourc_orient["reflections_details"]
    assert len(refls) == 2
    assert "wavelength" in refls[0]
    assert refls[0]["wavelength"] == 1.54


def test_list_orientation_runs(cat, RE, fourc, kappa):
    det = hw().noisy_det

    def scans():
        yield from bp.count([det])
        yield from bp.count([fourc])
        yield from bp.count([kappa])
        yield from bp.count([fourc, kappa])

    RE(scans())
    runs = hkl.util.list_orientation_runs(cat)
    # four sets of orientation info
    # (last scan has 2, first scan has none)
    assert len(runs.scan_id) == 4
    assert 1 not in runs.scan_id.to_list()  # no orientation
    assert runs.scan_id.to_list() == [2, 3, 4, 4]
    assert runs.diffractometer_name.to_list() == "fourc kappa fourc kappa".split()


def test_restore_orientation(cat, RE, fourc):
    RE(bp.count([fourc]))
    fourc_orient = hkl.util.run_orientation_info(cat[-1])["fourc"]

    e4cv = Fourc("", name="e4cv")
    assert len(e4cv.calc._samples) == 1

    # typical case : good match, restores successfully
    hkl.util.restore_orientation(fourc_orient, e4cv)
    assert len(e4cv.calc._samples) == 2
    sample = e4cv.calc.sample
    assert sample.name == "Si"
    assert len(sample.reflections) == 2
    numpy.testing.assert_array_equal(sample.reflections, [[4, 0, 0], [0, 4, 0]])
    # fmt: off
    numpy.testing.assert_array_almost_equal(
        fourc.calc.sample.UB,
        e4cv.calc.sample.UB,
    )
    # fmt: on

    # geometry mismatch, cannot restore
    k4cv = Kappa("", name="k4cv")
    assert len(k4cv.calc._samples) == 1
    assert len(k4cv.calc.sample.reflections) == 0
    with pytest.raises(ValueError) as exinfo:
        hkl.util.restore_orientation(fourc_orient, k4cv)
    args = exinfo.value.args
    assert len(args) == 1
    expected = "Geometries do not match: Orientation=E4CV, k4cv=K4CV, will not restore."
    assert args[0] == expected

    # different class name, restores successfully
    e4cv = SimulatedE4CV("", name="e4cv")
    assert len(e4cv.calc._samples) == 1
    assert len(e4cv.calc.sample.reflections) == 0
    hkl.util.restore_orientation(fourc_orient, e4cv)
    assert len(e4cv.calc._samples) == 2
    assert len(e4cv.calc.sample.reflections) == 2
    # fmt: off
    numpy.testing.assert_array_almost_equal(
        fourc.calc.sample.UB,
        e4cv.calc.sample.UB,
    )
    # fmt: on

    # different real axis names, restores successfully
    e4cv = Fourc("", name="e4cv")
    # change the real axis names
    e4cv.calc.physical_axis_names = {
        "omega": "john",
        "chi": "paul",
        "phi": "george",
        "tth": "ringo",
    }
    assert len(e4cv.calc._samples) == 1
    assert len(e4cv.calc.sample.reflections) == 0
    hkl.util.restore_orientation(fourc_orient, e4cv)
    assert len(e4cv.calc._samples) == 2
    assert len(e4cv.calc.sample.reflections) == 2
    # fmt: off
    numpy.testing.assert_array_almost_equal(
        fourc.calc.sample.UB,
        e4cv.calc.sample.UB,
    )
    # fmt: on


def test_restore_sample(cat, RE, fourc):
    RE(bp.count([fourc]))
    fourc_orient = hkl.util.run_orientation_info(cat[-1])["fourc"]

    # Add sample to new diffractometer.
    e4cv = Fourc("", name="e4cv")
    hkl.util.restore_sample(fourc_orient, e4cv)
    assert len(e4cv.calc._samples) == 2
    assert len(e4cv.calc.sample.reflections) == 0
    assert e4cv.calc.sample.name == fourc.calc.sample.name

    # Sample already defined, will not make another.
    with pytest.raises(ValueError) as exinfo:
        hkl.util.restore_sample(fourc_orient, e4cv)
    expected = "Sample 'Si' already exists in e4cv."
    assert exinfo.value.args[0] == expected



import gi
from ophyd import Component as Cpt
from ophyd import PseudoSingle

gi.require_version("Hkl", "5.0")
from hkl import SimulatedE4CV


class Fourc(SimulatedE4CV):
    ...


FOURC_SETUP_CODE = """
from ophyd import Component as Cpt
from ophyd import SoftPositioner
import gi
gi.require_version('Hkl', '5.0')
from hkl import SimulatedE4CV

class Fourc(SimulatedE4CV):
    ...
"""


def test_plain_fourc_not_fatal(testdir):
    test_code = FOURC_SETUP_CODE
    test_code += "\n" + "fourc = Fourc('', name='fourc')"
    testdir.makepyfile(test_code)
    result = testdir.runpytest_subprocess()
    result.stderr.no_fnmatch_line("*Fatal Python error*")


def test_extra_real_but_not_fatal(testdir):
    test_code = FOURC_SETUP_CODE
    test_code += "\n" + "class FourcSub(Fourc):"
    test_code += "\n" + "    extra = Cpt(SoftPositioner, init_pos=0)"
    test_code += "\n" + "fourc = FourcSub('', name='fourc')"
    testdir.makepyfile(test_code)
    result = testdir.runpytest_subprocess()
    # result.stderr.no_fnmatch_line("*Fatal Python error*")
    result.stderr.str().startswith("Fatal Python error: Aborted")


def test_extra_real_not_fatal(testdir):
    test_code = FOURC_SETUP_CODE
    test_code += "\n" + "class FourcSub(Fourc):"
    test_code += "\n" + "    _real = ['omega', 'chi', 'phi', 'tth', ]"
    test_code += "\n" + "    extra = Cpt(SoftPositioner, init_pos=0)"
    test_code += "\n" + "fourc = FourcSub('', name='fourc')"
    testdir.makepyfile(test_code)
    result = testdir.runpytest_subprocess()
    result.stderr.no_fnmatch_line("*Fatal Python error*")


def test_extra_pseudo_not_fatal(testdir):
    test_code = FOURC_SETUP_CODE + "\n"
    test_code += "class FourcSub(Fourc):" + "\n"
    test_code += "    _pseudo = ['h', 'k', 'l', ]" + "\n"
    test_code += "    extra = Cpt(PseudoSingle, '')" + "\n"
    test_code += "fourc = FourcSub('', name='fourc')" + "\n"
    test_code += "print(fourc.position)" + "\n"
    test_code += "print(fourc.extra.position)" + "\n"
    testdir.makepyfile(test_code)
    result = testdir.runpytest_subprocess()
    result.stderr.no_fnmatch_line("*Fatal Python error*")


def test_fourc_position():
    fourc = Fourc("", name="fourc")
    assert fourc.position == (0, 0, 0)


def test_fourc_extra_pseudo():
    class FourcSub(Fourc):
        _pseudo = "h k l".split()
        p_extra = Cpt(PseudoSingle, "")

    fourc = FourcSub("", name="fourc")
    assert fourc.position == (0, 0, 0)
    assert hasattr(fourc.p_extra, "_idx")
    # TODO: this feature is broken in ophyd at this time
    # https://github.com/bluesky/hklpy/issues/48
    # https://github.com/bluesky/ophyd/issues/924
    assert fourc.p_extra._idx is None



from ophyd import Component as Cpt
from ophyd import SoftPositioner
import gi
import numpy as np
import numpy.testing
import pytest

gi.require_version("Hkl", "5.0")
# NOTE: MUST call gi.require_version() BEFORE import hkl
import hkl.calc
from hkl import Constraint
from hkl import E6C, SimMixin
from hkl import Lattice


TARDIS_TEST_MODE = "lifting_detector_mu"


class Tardis(SimMixin, E6C):
    # theta
    theta = Cpt(SoftPositioner, init_pos=0)
    omega = Cpt(SoftPositioner, init_pos=0)
    chi = Cpt(SoftPositioner, init_pos=0)
    phi = Cpt(SoftPositioner, init_pos=0)
    # delta, gamma
    delta = Cpt(SoftPositioner, init_pos=0)
    gamma = Cpt(SoftPositioner, init_pos=0)


@pytest.fixture(scope="function")
def tardis():
    tardis = Tardis("", name="tardis")
    tardis.calc.engine.mode = TARDIS_TEST_MODE
    # re-map Tardis' axis names onto what an E6C expects
    tardis.calc.physical_axis_names = {
        "mu": "theta",
        "omega": "omega",
        "chi": "chi",
        "phi": "phi",
        "gamma": "delta",
        "delta": "gamma",
    }
    tardis.wait_for_connection()
    return tardis


@pytest.fixture(scope="function")
def kcf_sample(tardis):
    # note: orientation matrix (below) was pre-computed with this wavelength
    # wavelength units must match lattice unit cell length units
    tardis.calc.wavelength = 13.317314715359827
    print("calc.wavelength is", tardis.calc.wavelength)
    print("sample is", tardis.calc.sample)
    print("position is", tardis.position)

    # lengths must have same units as wavelength
    # angles are in degrees
    lattice = Lattice(a=5.857, b=5.857, c=7.849, alpha=90.0, beta=90.0, gamma=90.0)

    # add the sample to the calculation engine
    tardis.calc.new_sample("KCF", lattice=lattice)

    # We can alternatively set the wavelength
    # (or photon energy) on the Tardis.calc instance.
    p1 = tardis.calc.Position(
        # fmt: off
        theta=48.42718305024724,
        omega=0.0,
        chi=0.0,
        phi=0.0,
        delta=115.65436271083637,
        gamma=3.0000034909999993,
        # fmt: on
    )
    r1 = tardis.calc.sample.add_reflection(0, 0, 1, position=p1)

    p2 = tardis.calc.Position(
        # fmt: off
        theta=138.42718305024724,
        omega=0.0,
        chi=0.0,
        phi=0.0,
        delta=115.65436271083637,
        gamma=3.0000034909999993,
        # fmt: on
    )
    r2 = tardis.calc.sample.add_reflection(1, 1, 0, position=p2)
    tardis.calc.sample.compute_UB(r1, r2)

    print("sample name is", tardis.sample_name.get())
    print("u matrix is", tardis.U.get(), tardis.U.describe())
    print("ub matrix is", tardis.UB.get(), tardis.UB.describe())
    print(
        # fmt: off
        "reflections:",
        tardis.reflections.get(),
        tardis.reflections.describe(),
        # fmt: on
    )
    print("ux is", tardis.ux.get(), tardis.ux.describe())
    print("uy is", tardis.uy.get(), tardis.uy.describe())
    print("uz is", tardis.uz.get(), tardis.uz.describe())
    print("lattice is", tardis.lattice.get(), tardis.lattice.describe())
    print(tardis.read())
    return kcf_sample


@pytest.fixture(scope="function")
def constrain(tardis):
    tardis.apply_constraints(
        dict(
            theta=Constraint(-181, 181, 0, True),
            # we don't have these!! Fix to 0
            phi=Constraint(0, 0, 0, False),
            chi=Constraint(0, 0, 0, False),
            omega=Constraint(0, 0, 0, False),
            # NOTE: Tardis detector stage names are swapped from canonical!
            delta=Constraint(-5, 180, 0, True),
            gamma=Constraint(-5, 180, 0, True),
        )
    )


def test_params(tardis):
    """
    Make sure the parameters are set correctly
    """
    calc = tardis.calc
    assert calc.pseudo_axis_names == "h k l".split()
    assert tuple(calc.physical_axis_names) == tardis.real_positioners._fields
    assert tardis.real_positioners._fields == tuple("theta omega chi phi delta gamma".split())
    assert calc.engine.mode == TARDIS_TEST_MODE

    # gamma
    calc["gamma"].limits = (-5, 180)
    calc["gamma"].value = 10
    calc["gamma"].fit = False

    assert calc["gamma"].limits == (-5, 180)
    assert calc["gamma"].value == 10
    assert calc["gamma"].fit is False

    # try another random set of parameters
    # in case it was initialized to the parameters we "set"
    calc["gamma"].limits = (-10, 180)
    calc["gamma"].value = 20
    calc["gamma"].fit = True

    assert calc["gamma"].limits == (-10, 180)
    assert calc["gamma"].value == 20
    assert calc["gamma"].fit is True


def test_reachable(tardis, kcf_sample, constrain):
    ppos = (0, 0, 1.1)
    rpos = (
        101.56806493825435,
        0.0,
        0.0,
        0.0,
        42.02226419522791,
        176.69158155966787,
    )

    tardis.move(ppos)
    print("tardis position is", tardis.position)
    print("tardis position is", tardis.calc.physical_positions)
    numpy.testing.assert_almost_equal(tardis.position, ppos)
    numpy.testing.assert_almost_equal(tardis.calc.physical_positions, rpos)


def test_inversion(tardis, kcf_sample, constrain):
    ppos = (0, 0, 1.1)
    rpos = (
        101.56806493825435,
        0.0,
        0.0,
        0.0,
        42.02226419522791,
        # invert gamma for this test:
        -176.69158155966787,
    )

    tardis.calc.inverted_axes = ["gamma"]
    tardis.calc.physical_positions = rpos

    assert not tardis.calc["omega"].inverted
    gamma = tardis.calc["gamma"]
    assert gamma.inverted
    numpy.testing.assert_almost_equal(gamma.limits, (-180.0, 5.0))  # inverted from (-5, 180)
    gamma.limits = (-180.0, 5.0)
    numpy.testing.assert_almost_equal(gamma.limits, (-180.0, 5.0))  # inverted from (-5, 180)

    numpy.testing.assert_almost_equal(tardis.calc.physical_positions, rpos)
    numpy.testing.assert_almost_equal(tardis.calc.inverse(rpos), ppos)


def test_unreachable(tardis, kcf_sample):
    print("position is", tardis.position)
    with pytest.raises(hkl.calc.UnreachableError) as exinfo:
        tardis.move((0, 0, 1.8))

    ex = exinfo.value
    print("grabbing last valid position", ex.pseudo, ex.physical)
    print("tardis position is", tardis.position)
    print("tardis position is", tardis.calc.physical_positions)

    # it should not have moved:
    numpy.testing.assert_almost_equal(tardis.position, (0, 0, 0))
    numpy.testing.assert_almost_equal(tardis.calc.physical_positions, [0] * 6)


@pytest.mark.parametrize(
    "theta, delta, gamma, h, k, l",
    [
        # original table, cannot compute inverse() to reasonable precision
        # (0.000, 0.000, 0.000, 0.000, 0.000, 0.000),
        # (0.075, 0.125, 0.000, -0.006, 0.013, 0.000),
        # (0.150, 0.250, 0.000, -0.013, 0.026, 0.000),
        # (0.225, 0.375, 0.000, -0.019, 0.038, 0.000),
        # (0.300, 0.500, 0.000, -0.025, 0.051, 0.000),
        # table recomputed using this UB matrix
        (0.000, 0.000, 0.0, 0.000000, 0.000000, 0.000000),
        (0.075, 0.125, 0.0, 0.000053, 0.000083, 0.000780),
        (0.150, 0.250, 0.0, 0.000107, 0.000166, 0.001561),
        (0.225, 0.375, 0.0, 0.000160, 0.000249, 0.002341),
        (0.300, 0.500, 0.0, 0.000214, 0.000333, 0.003121),
    ],
)
def test_issue62(tardis, kcf_sample, constrain, theta, delta, gamma, h, k, l):
    # simulate the scan, computing hkl from angles
    # RE(scan([hw.det, tardis], tardis.theta, 0, 0.3, tardis.delta, 0, 0.5, num=5))
    # values as reported from LiveTable

    tardis.energy_units.put("eV")
    tardis.energy.put(573)
    tardis.calc["gamma"].limits = (-2.81, 183.1)
    assert round(tardis.calc.energy, 5) == 0.573

    assert tardis.max_forward_iterations.get() == 100
    tardis.max_forward_iterations.put(500)  # increase from the default

    # test inverse() on each row in the table
    # fmt: off
    ref = tardis.inverse(
        theta=theta, omega=0, chi=0, phi=0, delta=delta, gamma=gamma
    )
    # fmt: on
    result = (ref.h, ref.k, ref.l)
    expected = (h, k, l)
    numpy.testing.assert_almost_equal(result, expected, decimal=6, err_msg=f"result={result}  expected={expected}")


# =========================================================
# new unit tests based on original tardis notebook
# Note: test_sample1() and test_sample1_calc_only() are nearly identical.
# They differ in that the first tests with a Diffractometer subclass
# (which provides for axis name remapping and energy units) while
# the second only tests the calc and below (canonical axis names and
# energy in keV).


@pytest.fixture(scope="function")
def sample1(tardis):
    # test with remapped names, not canonical names

    # lattice cell lengths are in Angstrom, angles are in degrees
    a = 9.069
    c = 10.390
    tardis.calc.new_sample("sample1", lattice=(a, a, c, 90, 90, 120))

    tardis.energy_offset.put(0)
    tardis.energy_units.put("keV")
    tardis.energy.put(hkl.calc.A_KEV / 1.61198)

    pos_330 = tardis.calc.Position(delta=64.449, theta=25.285, chi=0.0, phi=0.0, omega=0.0, gamma=-0.871)
    pos_520 = tardis.calc.Position(delta=79.712, theta=46.816, chi=0.0, phi=0.0, omega=0.0, gamma=-1.374)

    r_330 = tardis.calc.sample.add_reflection(3, 3, 0, position=pos_330)
    r_520 = tardis.calc.sample.add_reflection(5, 2, 0, position=pos_520)

    tardis.calc.sample.compute_UB(r_330, r_520)

    def constrain(axis_name, low, high, value, fit):
        axis = tardis.calc[axis_name]
        axis.limits = (low, high)
        axis.value = value
        axis.fit = fit

    # use geometry's canonical names with calc, not the remapped names
    constrain("mu", -181, 181, 0, True)  # Tardis name: theta
    constrain("gamma", -5, 180, 0, True)  # Tardis name: delta
    constrain("delta", -5, 180, 0, True)  # Tardis name: gamma
    # Do not have these axes, fix to 0.
    for ax in "phi chi omega".split():
        constrain(ax, 0, 0, 0, False)


def test_sample1(sample1, tardis):
    sample = tardis.calc.sample
    wavelength = 1.61198
    assert sample is not None
    assert sample.name == "sample1"
    assert abs(tardis.calc.wavelength - wavelength) < 1e-6
    assert len(sample.reflections) == 2
    assert sample.reflections == [(3, 3, 0), (5, 2, 0)]

    # NOTE: reflections_details() uses canonical names for real positioners
    assert len(sample.reflections_details) == 2
    assert sample.reflections_details[0] == dict(
        # sorted alphabetically
        reflection=dict(h=3.0, k=3.0, l=0.0),
        flag=1,
        wavelength=wavelength,
        position=dict(
            # note: reflections_details uses canonical names, not remapped names!
            chi=0.0,
            delta=-0.871,  # Tardis: gamma
            gamma=64.449,  # Tardis: delta
            mu=25.285,  # Tardis: theta
            omega=0.0,
            phi=0.0,
        ),
        orientation_reflection=True,
    )
    assert sample.reflections_details[1] == dict(
        # sorted alphabetically
        reflection=dict(h=5.0, k=2.0, l=0.0),
        flag=1,
        wavelength=wavelength,
        position=dict(
            # note: reflections_details uses canonical names, not remapped names!
            chi=0.0,
            delta=-1.374,  # Tardis: gamma
            gamma=79.712,  # Tardis: delta
            mu=46.816,  # Tardis: theta
            omega=0.0,
            phi=0.0,
        ),
        orientation_reflection=True,
    )
    numpy.testing.assert_almost_equal(
        sample.UB,
        np.array(
            [
                [0.31323551, -0.4807593, 0.01113654],
                [0.73590724, 0.63942704, 0.01003773],
                [-0.01798898, -0.00176066, 0.60454803],
            ]
        ),
        7,
    )


def test_sample1_calc_only():
    # These comparisons start with the Tardis' calc support (no Diffractometer object)

    tardis_calc = hkl.calc.CalcE6C()

    assert tardis_calc.engine.mode == "bissector_vertical"
    tardis_calc.engine.mode = TARDIS_TEST_MODE
    assert tardis_calc.engine.mode != "bissector_vertical"

    assert tardis_calc.wavelength == 1.54
    tardis_calc.wavelength = 1.61198
    assert tardis_calc.wavelength != 1.54
    assert abs(tardis_calc.energy - 7.69142297) < 5e-7

    # lattice cell lengths are in Angstrom, angles are in degrees
    a = 9.069
    c = 10.390
    sample = tardis_calc.new_sample("sample1", lattice=(a, a, c, 90, 90, 120))
    assert sample.name == "sample1"

    pos_330 = tardis_calc.Position(gamma=64.449, mu=25.285, chi=0.0, phi=0.0, omega=0.0, delta=-0.871)
    pos_520 = tardis_calc.Position(gamma=79.712, mu=46.816, chi=0.0, phi=0.0, omega=0.0, delta=-1.374)

    r_330 = sample.add_reflection(3, 3, 0, position=pos_330)
    r_520 = sample.add_reflection(5, 2, 0, position=pos_520)
    assert len(sample.reflections) == 2

    UB = tardis_calc.sample.compute_UB(r_330, r_520)
    expected = np.array(
        [
            [0.31323551, -0.4807593, 0.01113654],
            [0.73590724, 0.63942704, 0.01003773],
            [-0.01798898, -0.00176066, 0.60454803],
        ]
    )
    abs_diff = np.abs(UB - expected)
    numpy.testing.assert_array_less(abs_diff, 1e-3)
    numpy.testing.assert_allclose(UB, expected, 7, verbose=True)


def test_sample1_UB(sample1, tardis):
    # test the sample1 lattice and orientation
    numpy.testing.assert_almost_equal(
        tardis.calc.sample.UB,
        np.array(
            [
                [0.31323551, -0.4807593, 0.01113654],
                [0.73590724, 0.63942704, 0.01003773],
                [-0.01798898, -0.00176066, 0.60454803],
            ]
        ),
        7,
    )


@pytest.mark.parametrize(
    "wavelength, h, k, l, angles",
    [
        # lambda  h  k  l   delta    theta   chi phi omega gamma
        (1.61198, 4, 4, 0, [90.6303, 38.3762, 0, 0, 0, -1.1613]),
        (1.61198, 4, 1, 0, [56.0970, 40.2200, 0, 0, 0, -1.0837]),
        (1.60911, 6, 0, 0, [75.8452, 60.9935, 0, 0, 0, -1.5840]),
        (1.60954, 3, 2, 0, [53.0521, 26.1738, 0, 0, 0, -0.8438]),
        (1.60954, 5, 4, 0, [106.3205, 49.8923, 0, 0, 0, -1.4237]),
        (1.60954, 4, 5, 0, [106.3189, 42.5493, 0, 0, 0, -1.1854]),
    ],
)
def test_tardis_forward(sample1, tardis, wavelength, h, k, l, angles):
    # Experimentally found reflections @ Lambda = 1.61198 A
    # (4, 4, 0) = [90.628, 38.373, 0, 0, 0, -1.156]
    # (4, 1, 0) = [56.100, 40.220, 0, 0, 0, -1.091]
    # @ Lambda = 1.60911
    # (6, 0, 0) = [75.900, 61.000, 0, 0, 0, -1.637]
    # @ Lambda = 1.60954
    # (3, 2, 0) = [53.090, 26.144, 0, 0, 0, -.933]
    # (5, 4, 0) = [106.415, 49.900, 0, 0, 0, -1.535]
    # (4, 5, 0) = [106.403, 42.586, 0, 0, 0, -1.183]
    # Note angles these differ slightly from the test cases.
    # The experimental values will only pass if tolerance is increased to 0.12 !!

    # move the first angle into proper position
    angles.insert(-1, angles.pop(0))
    tolerance = 0.0001

    tardis.calc.wavelength = wavelength
    result = tardis.forward((h, k, l))
    for i, ax_name in enumerate(result._fields):
        axis = getattr(result, ax_name)
        expected = angles[i]
        assert abs(axis - expected) <= tolerance, f"({h} {k} {l}) {ax_name}"



import gi
import numpy.testing
import pint
import pyRestTable
import pytest

gi.require_version("Hkl", "5.0")
# NOTE: MUST call gi.require_version() BEFORE import hkl
from hkl.calc import A_KEV
from hkl.diffract import Constraint
from hkl import SimulatedE4CV


class Fourc(SimulatedE4CV):
    ...


@pytest.fixture(scope="function")
def fourc():
    fourc = Fourc("", name="fourc")
    fourc.wait_for_connection()
    fourc._update_calc_energy()
    return fourc


def test_calc_energy_permit(fourc):
    assert fourc._calc_energy_update_permitted
    fourc.energy_update_calc_flag.put(False)
    assert not fourc._calc_energy_update_permitted

    nrg = fourc.calc.energy
    fourc.energy.put(5.989)  # BTW: Cr K absorption edge
    numpy.testing.assert_almost_equal(fourc.energy.get(), 5.989)
    numpy.testing.assert_almost_equal(fourc.calc.energy, nrg)

    fourc._energy_changed()
    numpy.testing.assert_almost_equal(fourc.calc.energy, nrg)

    fourc._energy_changed(fourc.energy.get())
    numpy.testing.assert_almost_equal(fourc.calc.energy, nrg)

    fourc._energy_changed(5.989)
    numpy.testing.assert_almost_equal(fourc.calc.energy, nrg)

    fourc._update_calc_energy()
    numpy.testing.assert_almost_equal(fourc.calc.energy, 5.989)

    # test that value argument is ignored
    fourc._update_calc_energy(A_KEV / 1)
    numpy.testing.assert_almost_equal(fourc.calc.energy, 5.989)


def test_energy(fourc):
    numpy.testing.assert_almost_equal(fourc.energy.get(), fourc.calc.energy)

    for nrg in (8.0, 8.04, 9.0, 0.931):
        fourc.energy.put(nrg)
        numpy.testing.assert_almost_equal(fourc.energy.get(), nrg)
        numpy.testing.assert_almost_equal(fourc.calc.energy, nrg)
        numpy.testing.assert_almost_equal(fourc.calc.wavelength, A_KEV / nrg)


def test_energy_offset(fourc):
    assert fourc.energy_offset.get() == 0

    nrg = 8.0
    fourc.energy.put(nrg)
    numpy.testing.assert_almost_equal(fourc.energy.get(), nrg)
    numpy.testing.assert_almost_equal(fourc.energy.get(), fourc.calc.energy)

    for offset in (0.05, -0.1):
        fourc.energy_offset.put(offset)
        fourc.energy.put(nrg)
        numpy.testing.assert_almost_equal(fourc.energy.get(), nrg)
        numpy.testing.assert_almost_equal(fourc.energy.get() + offset, fourc.calc.energy)


def test_energy_offset_units(fourc):
    assert fourc.energy_offset.get() == 0
    assert fourc.energy_units.get() == "keV"
    fourc.energy_units.put("eV")
    assert fourc.energy_units.get() == "eV"

    nrg = 931
    fourc.energy.put(nrg)
    numpy.testing.assert_almost_equal(fourc.energy.get(), nrg)
    numpy.testing.assert_almost_equal(fourc.energy.get() / 1000, fourc.calc.energy)

    for offset in (5, -6):
        fourc.energy_offset.put(offset)
        fourc.energy.put(nrg)
        numpy.testing.assert_almost_equal(fourc.energy.get(), nrg)
        numpy.testing.assert_almost_equal((fourc.energy.get() + offset) / 1000, fourc.calc.energy)


def test_energy_units_931eV(fourc):
    assert fourc.energy_units.get() == "keV"
    fourc.energy_units.put("eV")
    assert fourc.energy_units.get() == "eV"

    eV = 931
    fourc.energy.put(eV)
    numpy.testing.assert_almost_equal(fourc.energy.get(), eV)
    numpy.testing.assert_almost_equal(fourc.calc.energy, eV / 1000)


def test_energy_units_issue79(fourc):
    # issue #79
    fourc.energy_units.put("eV")
    fourc.energy_offset.put(0)
    eV = 1746
    fourc.energy.put(eV)
    numpy.testing.assert_almost_equal(fourc.calc.energy, eV / 1000)
    numpy.testing.assert_almost_equal(
        # fmt: off
        pint.Quantity(fourc.calc.energy, "keV").to(fourc.energy_units.get()).magnitude,
        fourc.energy.get(),
        # fmt: on
    )


def test_energy_units_offset(fourc):
    fourc.energy_units.put("keV")
    fourc.energy.put(7.985)
    fourc.energy_offset.put(0.015)
    assert fourc.calc.energy == 8.0
    assert round(fourc.energy.get(), 6) == 7.985

    fourc.energy.put(8)
    assert fourc.calc.energy == 8.015
    assert round(fourc.energy.get(), 6) == 8

    fourc.energy_offset.put(0.0)
    assert fourc.calc.energy == 8.0


def test_energy_units_issue86(fourc):
    # issue #86
    # changing units or offset changes .energy, not .calc.energy
    fourc.energy.put(8)
    fourc.energy_offset.put(0.015)
    fourc.energy_units.put("eV")
    # test interim state when fourc.energy value has not changed but units have
    assert round(fourc.calc.energy, 6) == 8.015e-3
    assert round(fourc.energy.get(), 1) == 8

    fourc.energy.put(8000)
    assert round(fourc.calc.energy, 8) == 8.000015
    assert round(fourc.energy.get(), 1) == 8000

    fourc.energy_offset.put(15)
    assert round(fourc.calc.energy, 8) == 8.015
    assert round(fourc.energy.get(), 1) == 8000

    fourc.energy.put(8000)
    assert round(fourc.calc.energy, 8) == 8.015
    assert round(fourc.energy.get(), 1) == 8000


def test_names(fourc):
    assert fourc.geometry_name.get() == "E4CV"
    assert fourc.class_name.get() == "Fourc"


def test_forward_solutions_table(fourc):
    fourc.energy.put(A_KEV / 1.54)

    # (100) has chi ~ 0 which poses occasional roundoff errors
    # (sometimes -0.00000, sometimes 0.00000)
    sol = fourc.forward(1, 0, 0)
    assert pytest.approx(sol.omega, 1e-5) == -30
    assert pytest.approx(sol.chi, 1e-5) == 0
    assert pytest.approx(sol.phi, 1e-5) == -90
    assert pytest.approx(sol.tth, 1e-5) == -60

    tbl = fourc.forward_solutions_table(
        # fmt: off
        [
            [1, 1, 0],
            [1, 1, 1],
            [100, 1, 1],  # no solutions
        ]
        # fmt: on
    )
    received = str(tbl).splitlines()
    expected = [
        "=========== ======== ===== ======== ==== =====",
        "(hkl)       solution omega chi      phi  tth  ",
        "=========== ======== ===== ======== ==== =====",
        "[1, 1, 0]   0        45.0  45.0     90.0 90.0 ",
        "[1, 1, 1]   0        60.0  35.26439 45.0 120.0",
        "[100, 1, 1] none                              ",
        "=========== ======== ===== ======== ==== =====",
    ]
    for r, e in zip(received, expected):
        assert r == e


def test_pa(fourc, capsys):
    tbl = fourc.pa()
    assert isinstance(tbl, pyRestTable.Table)
    out, err = capsys.readouterr()
    assert len(out) > 0
    assert err == ""
    out = [v.rstrip() for v in out.strip().splitlines()]
    expected = [
        "===================== ====================================================================",
        "term                  value",
        "===================== ====================================================================",
        "diffractometer        fourc",
        "geometry              E4CV",
        "class                 Fourc",
        "energy (keV)          8.00000",
        "wavelength (angstrom) 1.54980",
        "calc engine           hkl",
        "mode                  bissector",
        "positions             ===== =======",
        "                      name  value",
        "                      ===== =======",
        "                      omega 0.00000",
        "                      chi   0.00000",
        "                      phi   0.00000",
        "                      tth   0.00000",
        "                      ===== =======",
        "constraints           ===== ========= ========== ===== ====",
        "                      axis  low_limit high_limit value fit",
        "                      ===== ========= ========== ===== ====",
        "                      omega -180.0    180.0      0.0   True",
        "                      chi   -180.0    180.0      0.0   True",
        "                      phi   -180.0    180.0      0.0   True",
        "                      tth   -180.0    180.0      0.0   True",
        "                      ===== ========= ========== ===== ====",
        "sample: main          ================ ===================================================",
        "                      term             value",
        "                      ================ ===================================================",
        "                      unit cell edges  a=1.54, b=1.54, c=1.54",
        "                      unit cell angles alpha=90.0, beta=90.0, gamma=90.0",
        "                      [U]              [[1. 0. 0.]",
        "                                        [0. 1. 0.]",
        "                                        [0. 0. 1.]]",
        "                      [UB]             [[ 4.07999046e+00 -2.49827363e-16 -2.49827363e-16]",
        "                                        [ 0.00000000e+00  4.07999046e+00 -2.49827363e-16]",
        "                                        [ 0.00000000e+00  0.00000000e+00  4.07999046e+00]]",
        "                      ================ ===================================================",
        "===================== ====================================================================",
    ]
    assert len(out) == len(expected)
    assert out == expected


def test_wh(fourc, capsys):
    tbl = fourc.wh()
    assert isinstance(tbl, pyRestTable.Table)
    out, err = capsys.readouterr()
    assert len(out) > 0
    assert err == ""
    out = [v.rstrip() for v in out.strip().splitlines()]
    expected = [
        "===================== ========= =========",
        "term                  value     axis_type",
        "===================== ========= =========",
        "diffractometer        fourc",
        "sample name           main",
        "energy (keV)          8.00000",
        "wavelength (angstrom) 1.54980",
        "calc engine           hkl",
        "mode                  bissector",
        "h                     0.0       pseudo",
        "k                     0.0       pseudo",
        "l                     0.0       pseudo",
        "omega                 0         real",
        "chi                   0         real",
        "phi                   0         real",
        "tth                   0         real",
        "===================== ========= =========",
    ]
    assert len(out) == len(expected)
    assert out == expected


def test_show_constraints(fourc, capsys):
    fourc.show_constraints()
    out, err = capsys.readouterr()
    assert len(out) > 0
    assert err == ""
    out = [v.rstrip() for v in out.strip().splitlines()]
    expected = [
        "===== ========= ========== ===== ====",
        "axis  low_limit high_limit value fit",
        "===== ========= ========== ===== ====",
        "omega -180.0    180.0      0.0   True",
        "chi   -180.0    180.0      0.0   True",
        "phi   -180.0    180.0      0.0   True",
        "tth   -180.0    180.0      0.0   True",
        "===== ========= ========== ===== ====",
    ]
    for r, e in zip(out, expected):
        assert r.rstrip() == e.rstrip()


def test_apply_constraints(fourc):
    fourc.energy.put(A_KEV / 1.54)
    # fmt: off
    fourc.apply_constraints(
        {
            "tth": Constraint(0, 180, 0, True),
            "chi": Constraint(0, 180, 0, True),
        }
    )
    # fmt: on
    sol = fourc.forward(1, 0, 0)
    assert pytest.approx(sol.omega, 1e-5) == 30
    assert pytest.approx(sol.chi, 1e-5) == 0
    assert pytest.approx(sol.phi, 1e-5) == 90
    assert pytest.approx(sol.tth, 1e-5) == 60


def test_specify_engine():
    import hkl
    import numpy as np
    from ophyd import Component as Cpt
    from ophyd import PseudoSingle
    from ophyd import SoftPositioner

    class Q4C(hkl.E4CV):
        q = Cpt(PseudoSingle, "")
        omega = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0)
        chi = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0)
        phi = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0)
        tth = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0)

    q4c = Q4C("", name="q4c")
    assert q4c.calc.engine.name == "hkl"

    q4c = Q4C("", name="q4c", engine="q")
    assert q4c.calc.engine.name == "q"
    q = 1.0
    angle = 2 * np.arcsin(q * q4c.calc.wavelength / 4 / np.pi) * 180 / np.pi
    value = q4c.forward(q)
    assert round(value.tth, 5) == round(angle, 5)
    assert value.omega == 0.0
    assert value.chi == 0.0
    assert value.phi == 0.0



import epics
import logging
import unittest


logger = logging.getLogger("ophyd_session_test")


def setup_package():
    pass


def teardown_package():
    pass


def main(is_main):
    # fmt = '%(asctime)-15s [%(levelname)s] %(message)s'
    # logging.basicConfig(format=fmt, level=logging.DEBUG)
    epics.ca.use_initial_context()

    logger = logging.getLogger("logger")
    logger.setLevel(logging.INFO)

    if is_main:
        setup_package()
        unittest.main()
        teardown_package()



import gi
import numpy as np
import pytest

gi.require_version("Hkl", "5.0")
# NOTE: MUST call gi.require_version() BEFORE import hkl
from hkl import SimulatedE4CV


class Fourc(SimulatedE4CV):
    ...


@pytest.fixture(scope="function")
def fourc():
    fourc = Fourc("", name="fourc")
    return fourc


def test_compute_UB(fourc):
    sample = fourc.calc.sample
    default_UB = sample.UB

    fourc.calc.wavelength = 1.54
    #  compute the UB matrix from two reflections
    r1 = sample.add_reflection(-1, 0, 0, (30, 0, -90, 60))
    r2 = sample.add_reflection(0, 1, 1, (45, 45, 0, 90))
    sample.compute_UB(r1, r2)
    np.testing.assert_array_almost_equal(default_UB, sample.UB)

    #  compute the UB matrix from two other reflections
    # fmt: off
    r3 = sample.add_reflection(
        0.1, 0.2, 0.3,
        (10.7826, 32.3115, 18.4349, 21.5652)
    )
    r4 = sample.add_reflection(
        0.5, 0, 0.001,
        (14.4775, 0, 89.8854, 28.9551)
    )
    # fmt: on
    sample.compute_UB(r3, r4)
    # Since the angles are rounded to 4 decimal places,
    # limit the comparison as well.
    np.testing.assert_array_almost_equal(default_UB, sample.UB, decimal=5)


def test_orientation_reflections(fourc):
    sample = fourc.calc.sample
    assert len(sample._orientation_reflections) == 0

    fourc.calc.wavelength = 1.54
    r1 = sample.add_reflection(-1, 0, 0, (30, 0, -90, 60))
    r2 = sample.add_reflection(0, 1, 1, (45, 45, 0, 90))
    sample.compute_UB(r1, r2)
    assert len(sample._orientation_reflections) == 2
    details = sample.reflections_details
    assert details[0]["orientation_reflection"]
    assert details[1]["orientation_reflection"]

    # fmt: off
    r3 = sample.add_reflection(
        .1, .2, .3,
        (10.7826, 32.3115, 18.4349, 21.5652)
    )
    r4 = sample.add_reflection(
        .5, 0, .001,
        (14.4775, 0, 89.8854, 28.9551)
    )
    # fmt: on
    sample.compute_UB(r3, r4)
    assert len(sample._orientation_reflections) == 2
    details = sample.reflections_details
    assert not details[0]["orientation_reflection"]
    assert not details[1]["orientation_reflection"]
    assert details[2]["orientation_reflection"]
    assert details[3]["orientation_reflection"]

    sample.compute_UB(r1, r2)
    assert len(sample._orientation_reflections) == 2
    details = sample.reflections_details
    assert details[0]["orientation_reflection"]
    assert details[1]["orientation_reflection"]
    assert not details[2]["orientation_reflection"]
    assert not details[3]["orientation_reflection"]

    sample.remove_reflection(r3)
    assert len(sample._orientation_reflections) == 2
    sample.remove_reflection((0.5, 0, 0.001))
    assert len(sample._orientation_reflections) == 2


def test_reflections_details(fourc):
    def compare(expected, received):
        for e, r in list(zip(expected, received)):
            # scalar comparisons
            for key in "flag orientation_reflection wavelength".split():
                assert e[key] == r[key]
            # dictionary comparisons
            for key in "position reflection".split():
                for k, v in e[key].items():
                    assert round(abs(v - r[key][k])) == 0

    sample = fourc.calc.sample
    assert len(sample._orientation_reflections) == 0

    fourc.calc.wavelength = 1.54
    r1 = sample.add_reflection(-1, 0, 0, (30, 0, -90, 60))

    details = sample.reflections_details
    # fmt: off
    expected = [
        dict(
            flag=1,
            orientation_reflection=False,
            position=dict(omega=30, chi=0, phi=-90, tth=60),
            reflection=dict(h=-1, k=0, l=0),
            wavelength=1.54
        )
    ]
    # fmt: on
    compare(expected, details)

    r2 = sample.add_reflection(0, 1, 1, (45, 45, 0, 90))
    # fmt: off
    expected += [
        dict(
            flag=1,
            orientation_reflection=False,
            position=dict(omega=45, chi=45, phi=0, tth=90),
            reflection=dict(h=0, k=1, l=1),
            wavelength=1.54
        )
    ]
    sample.add_reflection(
        .1, .2, .3,
        (10.7826, 32.3115, 18.4349, 21.5652)
    )
    expected += [
        dict(
            flag=1,
            orientation_reflection=False,
            position=dict(
                omega=10.7826,
                chi=32.3115,
                phi=18.4349,
                tth=21.5652
            ),
            reflection=dict(h=.1, k=.2, l=.3),
            wavelength=1.54
        )
    ]
    sample.add_reflection(
        .5, 0, .001,
        (14.4775, 0, 89.8854, 28.9551)
    )
    expected += [
        dict(
            flag=1,
            orientation_reflection=False,
            position=dict(
                omega=14.4775,
                chi=0,
                phi=89.8854,
                tth=28.9551
            ),
            reflection=dict(h=.5, k=0, l=1e-3),
            wavelength=1.54
        ),
    ]
    # fmt: on
    details = sample.reflections_details
    compare(expected, details)

    sample.compute_UB(r1, r2)
    expected[0]["orientation_reflection"] = True
    expected[1]["orientation_reflection"] = True

    details = sample.reflections_details
    compare(expected, details)

    sample.UB = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    assert len(sample._orientation_reflections) == 0
    expected[0]["orientation_reflection"] = False
    expected[1]["orientation_reflection"] = False

    details = sample.reflections_details
    compare(expected, details)


def test_swap_orientation_reflections(fourc):
    sample = fourc.calc.sample

    fourc.calc.wavelength = 1.54
    #  compute the UB matrix from two reflections
    assert len(sample._orientation_reflections) == 0

    with pytest.raises(ValueError) as exinfo:
        sample.swap_orientation_reflections()
    expected = "Must have exactly 2 orientation reflections defined"
    assert expected in str(exinfo.value)

    r1 = sample.add_reflection(-1, 0, 0, (30, 0, -90, 60))
    assert len(sample._orientation_reflections) == 0

    r2 = sample.add_reflection(0, 1, 1, (45, 45, 0, 90))
    assert len(sample._orientation_reflections) == 0

    ub12 = sample.compute_UB(r1, r2)
    assert len(sample._orientation_reflections) == 2
    assert sample._orientation_reflections == [r1, r2]

    ub21 = sample.swap_orientation_reflections()
    assert len(sample._orientation_reflections) == 2
    assert sample._orientation_reflections == [r2, r1]
    assert (ub12 != ub21).any()

    r3 = sample.add_reflection(1, 1, 1, (1, 2, 3, 4))
    assert len(sample._orientation_reflections) == 2

    ubswap = sample.swap_orientation_reflections()
    assert len(sample._orientation_reflections) == 2
    assert sample._orientation_reflections == [r1, r2]
    assert (ub12 == ubswap).all()

    sample._orientation_reflections.append(r3)
    assert len(sample._orientation_reflections) == 3
    assert sample._orientation_reflections == [r1, r2, r3]
    with pytest.raises(ValueError) as exinfo:
        sample.swap_orientation_reflections()
    expected = "Must have exactly 2 orientation reflections defined"
    assert expected in str(exinfo.value)



"""
diffract
--------

Common Support for diffractometers

.. autosummary::

    ~Diffractometer

"""


from ophyd import Component as Cpt
from ophyd import PositionerBase
from ophyd import PseudoPositioner
from ophyd import Signal
from ophyd.pseudopos import pseudo_position_argument
from ophyd.pseudopos import real_position_argument
from ophyd.signal import AttributeSignal, ArrayAttributeSignal
import logging
import pint
import pyRestTable

from . import calc
from . import __version__
from .util import Constraint


__all__ = """
    Diffractometer
""".split()
logger = logging.getLogger(__name__)


class Diffractometer(PseudoPositioner):
    """Diffractometer pseudopositioner

    .. autosummary::

        ~calc
        ~engine
        ~forward
        ~inverse
        ~forward_solutions_table
        ~apply_constraints
        ~reset_constraints
        ~show_constraints
        ~undo_last_constraints
        ~pa
        ~wh
        ~_calc_energy_update_permitted
        ~_constraints_dict
        ~_constraints_for_databroker
        ~_energy_changed
        ~_energy_offset_changed
        ~_energy_units_changed
        ~_push_current_constraints
        ~_set_constraints
        ~_update_calc_energy


    This has a corresponding calculation engine from **hklpy** that does
    forward and inverse calculations.

    If instantiating a specific diffractometer class such as `E4C`, `E6C`,
    neither the `calc_inst` or the `calc_kw` parameters are required.

    However, there is the option to either pass in a calculation
    instance (with `calc_inst`) or keywords for the default calculation
    class (using `calc_kw`) to instantiate a new one.

    Parameters
    ----------
    prefix : str
        PV prefix for all components
    calc_kw : dict, optional
        Initializer arguments for the calc class
    decision_fcn : callable, optional
        The decision function to use when multiple solutions exist for a given
        forward calculation. Defaults to arbitrarily picking the first
        solution.
    engine : str, optional
        Calculation engine name.  Default: ``hkl``
    read_attrs : list, optional
        Read attributes default to all pseudo and real positioners
    configuration_attrs : list, optional
        Defaults to the UB matrix and energy
    parent : Device, optional
        Parent device
    name : str, optional
        Device name

    Attributes
    ----------
    calc_class : sub-class of CalcRecip
        Reciprocal calculation class used with this diffractometer. If ``None``
        (as used in `hkl.diffract.Diffractometer`), `calc_inst` must be
        specified upon initialization.

    See Also
    --------
    :class:`~hkl.geometries.E4CH`
    :class:`~hkl.geometries.E4CV`
    :class:`~hkl.geometries.E6C`
    :class:`~hkl.geometries.K4CV`
    :class:`~hkl.geometries.K6C`
    :class:`~hkl.geometries.Med2p3`
    :class:`~hkl.geometries.Petra3_p09_eh2`
    :class:`~hkl.geometries.SoleilMars`
    :class:`~hkl.geometries.SoleilSiriusKappa`
    :class:`~hkl.geometries.SoleilSiriusTurret`
    :class:`~hkl.geometries.SoleilSixs`
    :class:`~hkl.geometries.SoleilSixsMed1p2`
    :class:`~hkl.geometries.SoleilSixsMed2p2`
    :class:`~hkl.geometries.Zaxis`
    """

    calc_class = None

    # see: Documentation has examples to use an EPICS PV for energy.
    energy = Cpt(Signal, value=8.0, doc="Energy (in keV)")
    energy_units = Cpt(Signal, value="keV")
    energy_offset = Cpt(Signal, value=0)
    energy_update_calc_flag = Cpt(Signal, value=True)

    geometry_name = Cpt(
        # fmt: off
        AttributeSignal,
        attr="calc.geometry_name",
        doc="Diffractometer Geometry name",
        write_access=False,
        # fmt: on
    )
    class_name = Cpt(
        # fmt: off
        AttributeSignal,
        attr="__class__.__name__",
        doc="Diffractometer class name",
        write_access=False,
        # fmt: on
    )

    sample_name = Cpt(AttributeSignal, attr="calc.sample_name", doc="Sample name")
    lattice = Cpt(
        # fmt: off
        ArrayAttributeSignal,
        attr="calc.sample.lattice",
        doc="Sample lattice",
        # fmt: on
    )
    lattice_reciprocal = Cpt(
        # fmt: off
        AttributeSignal,
        attr="calc.sample.reciprocal",
        doc="Reciprocal lattice",
        # fmt: on
    )

    U = Cpt(AttributeSignal, attr="calc.sample.U", doc="U matrix")
    UB = Cpt(AttributeSignal, attr="calc.sample.UB", doc="UB matrix")
    # fmt: off
    reflections = Cpt(
        ArrayAttributeSignal,
        attr="calc.sample.reflections",
        doc="Reflections",
    )
    reflections_details = Cpt(
        AttributeSignal,
        attr="calc.sample.reflections_details",
        doc="Details of reflections",
    )
    ux = Cpt(
        AttributeSignal,
        attr="calc.sample.ux.value",
        doc="ux portion of the U matrix",
    )
    uy = Cpt(
        AttributeSignal,
        attr="calc.sample.uy.value",
        doc="uy portion of the U matrix",
    )
    uz = Cpt(
        AttributeSignal,
        attr="calc.sample.uz.value",
        doc="uz portion of the U matrix",
    )

    diffractometer_name = Cpt(
        AttributeSignal,
        attr="name",
        doc="Diffractometer name",
        write_access=False,
    )
    _hklpy_version_ = __version__
    _hklpy_version = Cpt(
        AttributeSignal,
        attr="_hklpy_version_",
        doc="hklpy version",
        write_access=False,
    )
    _pseudos = Cpt(
        AttributeSignal,
        attr="PseudoPosition._fields",
        doc="Pseudo Positioners",
        write_access=False,
    )
    _reals = Cpt(
        AttributeSignal,
        attr="RealPosition._fields",
        doc="Real Positioners",
        write_access=False,
    )
    _constraints = Cpt(
        ArrayAttributeSignal,
        attr="_constraints_for_databroker",
        doc="Constraints",
        write_access=False,
    )
    _mode = Cpt(
        AttributeSignal,
        attr="calc.engine.mode",
        doc="Mode of Operation",
        write_access=False,
    )
    orientation_attrs = Cpt(
        AttributeSignal,
        attr="_orientation_attrs",
        doc="Orientation Attributes",
        write_access=False,
    )

    _orientation_attrs = """
    orientation_attrs
    geometry_name class_name
    UB U ux uy uz
    energy energy_units energy_offset
    sample_name lattice lattice_reciprocal reflections_details
    _pseudos _reals
    _constraints _mode
    diffractometer_name _hklpy_version
    """.split()

    max_forward_iterations = Cpt(Signal, value=100, kind="config")
    # fmt: on

    # fmt: off
    def __init__(
        self,
        prefix,
        calc_kw=None,
        decision_fcn=None,
        calc_inst=None,
        engine="hkl",
        *,
        configuration_attrs=None,
        read_attrs=None,
        **kwargs,
    ):
        # fmt: on
        if calc_inst is not None:
            if not isinstance(calc_inst, self.calc_class):
                # fmt: off
                raise ValueError(
                    "Calculation instance must be derived"
                    f" from the class {self.calc_class}"
                )
            # fmt: on
            self._calc = calc_inst

        else:
            if calc_kw is None:
                calc_kw = {}

            self._calc = self.calc_class(engine=engine, lock_engine=True, **calc_kw)

        if not self.calc.engine_locked:
            # Reason for this is that the engine determines the pseudomotor
            # names, so if the engine is switched from underneath, the
            # pseudomotor will no longer function properly
            # fmt: off
            raise ValueError(
                "Calculation engine must be locked (CalcDiff.lock_engine set)"
            )
            # fmt: on

        if configuration_attrs is None:
            configuration_attrs = """
                UB energy reflections_details geometry_name class_name
            """.split()

        if decision_fcn is None:
            # the default decision function is to just grab solution #1:
            decision_fcn = calc.default_decision_function

        self._decision_fcn = decision_fcn

        super().__init__(
            # fmt: off
            prefix,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            **kwargs,
            # fmt: on
        )

        # write the crystal orientation information in descriptor doc
        self.configuration_attrs += self._orientation_attrs
        self._constraints_stack = []

        if read_attrs is None:
            # if unspecified, set the read attrs to the pseudo/real motor
            # positions once known
            self.read_attrs = list(self.PseudoPosition._fields) + list(self.RealPosition._fields)

        self.energy.subscribe(self._energy_changed, event_type=Signal.SUB_VALUE)
        self.energy_offset.subscribe(self._energy_offset_changed, event_type=Signal.SUB_VALUE)
        self.energy_units.subscribe(self._energy_units_changed, event_type=Signal.SUB_VALUE)

    @property
    def _calc_energy_update_permitted(self):
        """return boolean `True` if permitted"""
        acceptable_values = (1, "Yes", "locked", "OK", True, "On")
        return self.energy_update_calc_flag.get() in acceptable_values

    def _energy_changed(self, value=None, **kwargs):
        """
        Callback indicating that the energy signal was updated

        .. note::
            The `energy` signal is subscribed to this method
            in the :meth:`Diffractometer.__init__()` method.
        """
        if not self.connected:
            logger.warning(
                # fmt: off
                "%s not fully connected, %s.calc.energy not updated",
                self.name,
                self.name,
                # fmt: on
            )
            return

        if self._calc_energy_update_permitted:
            self._update_calc_energy()

    def _energy_offset_changed(self, value=None, **kwargs):
        """
        Callback indicating that the energy offset signal was updated.

        .. note::
            The ``energy_offset`` signal is subscribed to this method
            in the :meth:`Diffractometer.__init__()` method.
        """
        if not self.connected:
            logger.warning(
                # fmt: off
                "%s not fully connected, %s.calc.energy not updated",
                self.name,
                self.name,
                # fmt: on
            )
            return

        if self._calc_energy_update_permitted:
            self._update_calc_energy()

    def _energy_units_changed(self, value=None, **kwargs):
        """
        Callback indicating that the energy units signal was updated.

        .. note::
            The ``energy_units`` signal is subscribed to this method
            in the :meth:`Diffractometer.__init__()` method.
        """
        if not self.connected:
            logger.warning(
                # fmt: off
                "%s not fully connected, %s.calc.energy not updated",
                self.name,
                self.name,
                # fmt: on
            )
            return

        if self._calc_energy_update_permitted:
            self._update_calc_energy()

    def _update_calc_energy(self, value=None, **kwargs):
        """
        writes ``self.calc.energy`` from ``value`` or ``self.energy``.
        """
        if not self.connected:
            logger.warning(
                # fmt: off
                "%s not fully connected, %s.calc.energy not updated",
                self.name,
                self.name,
                # fmt: on
            )
            return

        value = float(self.energy.get())

        # energy_offset has same units as energy
        value += self.energy_offset.get()

        # comment these lines to skip unit conversion
        units = self.energy_units.get()
        if units != "keV":
            keV = pint.Quantity(value, units).to("keV")
            value = keV.magnitude

        if value <= 0:
            logger.debug("Computed energy(%s) is not positive", value)
            return
        logger.debug("setting %s.calc.energy = %s (keV)", self.name, value)
        self.calc.energy = value
        self._update_position()

    @property
    def calc(self):
        """The calculation instance"""
        return self._calc

    @property
    def engine(self):
        """The calculation engine associated with the diffractometer"""
        return self.calc.engine

    # TODO so these calculations change the internal state of the hkl
    # calculation class, which is probably not a good thing
    # -- it becomes a problem when someone uses these functions
    # outside of move()

    @pseudo_position_argument
    def forward(self, pseudo):
        """
        Calculate the real positions given the pseudo positions (hkl -> angles).

        Return the default solution using the ``_decision_fcn()``.
        """
        solutions = self.calc.forward_iter(
            start=self.position, end=pseudo, max_iters=self.max_forward_iterations.get()
        )
        logger.debug("pseudo to real: %s", solutions)
        return self._decision_fcn(pseudo, solutions)

    @real_position_argument
    def inverse(self, real):
        """
        Calculate the pseudo positions given the real positions (angles -> hkl).
        """
        self.calc.physical_positions = real
        return self.PseudoPosition(*self.calc.pseudo_positions)

    def check_value(self, pos):
        """
        Raise exception if pos is not within limits.

        In a scan, a subset of the pseudo axes may be directed,
        which are given in a dict from a set message from the
        bluesky RunEngine.

        It is not permitted to scan both pseudo and real positioners.
        """
        if isinstance(pos, dict):
            # Redefine and fill in any missing values.

            for axis, target in pos.items():
                if hasattr(self, axis):
                    p = getattr(self, axis)
                    if p in self.real_positioners:
                        p.check_value(target)
                else:
                    raise KeyError(f"{axis} not in {self.name}")

            pos = [pos.get(p.attr_name, p.position) for p in self.pseudo_positioners]
        super().check_value(pos)

    def apply_constraints(self, constraints):
        """
        Constrain the solutions of the diffractometer's forward() computation.

        This action will first save the current constraints onto
        a stack, enabling both *undo* and *reset* features.
        """
        self._push_current_constraints()
        self._set_constraints(constraints)

    def reset_constraints(self):
        """Set constraints back to initial settings."""
        if len(self._constraints_stack) > 0:
            self._set_constraints(self._constraints_stack[0])
            self._constraints_stack = []

    @property
    def _constraints_dict(self):
        """Return the constraints."""
        return {
            # fmt:off
            m: Constraint(
                *self.calc[m].limits,
                self.calc[m].value,
                self.calc[m].fit,
            )
            # fmt:on
            for m in self.RealPosition._fields
        }

    @property
    def _constraints_for_databroker(self):
        """
        Return the constraints for databroker.

        Cannot write a dictionary from bluesky because all values must
        be of the same data type, chosen from the first item in the
        list. Just write the values (the boolean will be written as a
        float.) The constraints will be written in the order of the real
        positioners.
        """
        # fmt: off
        return [
            tuple(self._constraints_dict[p]._asdict().values())
            for p in self.RealPosition._fields
        ]
        # fmt: on

    def show_constraints(self, fmt="simple", printing=True):
        """Print the current constraints in a table."""
        tbl = pyRestTable.Table()
        tbl.labels = "axis low_limit high_limit value fit".split()
        for k, c in self._constraints_dict.items():
            tbl.addRow((k, *c))

        if printing:
            print(tbl.reST(fmt=fmt))

        return tbl

    def undo_last_constraints(self):
        """
        Remove the current constraints additions, restore previous.
        """
        if len(self._constraints_stack) > 0:
            self._set_constraints(self._constraints_stack.pop())

    def _push_current_constraints(self):
        """push current constraints onto the stack"""
        constraints = {
            m: Constraint(*self.calc[m].limits, self.calc[m].value, self.calc[m].fit)
            for m in self.real_positioners._fields
        }
        self._constraints_stack.append(constraints)

    def _set_constraints(self, constraints):
        """set diffractometer's constraints"""
        for axis, constraint in constraints.items():
            self.calc[axis].limits = list(constraint)[0:2]
            #     constraint.low_limit,
            #     constraint.high_limit,
            # )
            self.calc[axis].value = constraint.value
            self.calc[axis].fit = constraint.fit

    def forward_solutions_table(self, reflections, full=False, digits=5):
        """
        Return table of computed solutions for each supplied (hkl) reflection.

        The solutions are calculated using the current UB matrix & constraints.

        Parameters
        ----------
        reflections : list of (h, k, l) reflections
            Each reflection is a tuple of 3 numbers,
            (h, k, l) of the reflection.
        full : bool
            If ``True``, show all solutions.  If ``False``,
            only show the default solution.
        digits : int
            Number of digits to roundoff each position
            value.  Default is 5.
        """
        _table = pyRestTable.Table()
        motors = self.real_positioners._fields
        _table.labels = "(hkl) solution".split() + list(motors)
        for reflection in reflections:
            try:
                solutions = self.calc.forward(reflection)
            except ValueError as exc:
                solutions = exc
            if isinstance(solutions, ValueError):
                row = [reflection, "none"]
                row += ["" for m in motors]
                _table.addRow(row)
            else:
                for i, s in enumerate(solutions):
                    row = [reflection, i]
                    row += [round(getattr(s, m), digits) for m in motors]
                    _table.addRow(row)
                    if not full:
                        break  # only show the first (default) solution
        return _table

    def pa(self, all_samples=False, printing=True):
        """
        Report (all) the diffractometer settings.

        EXAMPLE::

            In [1]: from hkl import SimulatedK4CV

            In [2]: k4cv = SimulatedK4CV('', name='k4cv')

            In [3]: k4cv.pa()  # FIXME lines are too long to include in source code
            ===================== ====================================================================
            term                  value
            ===================== ====================================================================
            diffractometer        k4cv
            geometry              K4CV
            class                 SimulatedK4CV
            energy (keV)          8.05092
            wavelength (angstrom) 1.54000
            calc engine           hkl
            mode                  bissector
            positions             ====== =======
                                name   value
                                ====== =======
                                komega 0.00000
                                kappa  0.00000
                                kphi   0.00000
                                tth    0.00000
                                ====== =======
            sample: main          ================ ===================================================
                                term             value
                                ================ ===================================================
                                unit cell edges  a=1.54, b=1.54, c=1.54
                                unit cell angles alpha=90.0, beta=90.0, gamma=90.0
                                [U]              [[1. 0. 0.]
                                                    [0. 1. 0.]
                                                    [0. 0. 1.]]
                                [UB]             [[ 4.07999046e+00 -2.49827363e-16 -2.49827363e-16]
                                                    [ 0.00000000e+00  4.07999046e+00 -2.49827363e-16]
                                                    [ 0.00000000e+00  0.00000000e+00  4.07999046e+00]]
                                ================ ===================================================
            ===================== ====================================================================

            Out[3]: <pyRestTable.rest_table.Table at 0x7f5c16503e20>

        """

        def addTable(tbl):
            return str(tbl).strip()

        def Package(**kwargs):
            return ", ".join([f"{k}={v}" for k, v in kwargs.items()])

        table = pyRestTable.Table()
        table.labels = "term value".split()

        table.addRow(("diffractometer", self.name))
        table.addRow(("geometry", self.calc._geometry.name_get()))
        table.addRow(("class", self.__class__.__name__))
        table.addRow(("energy (keV)", f"{self.calc.energy:.5f}"))
        table.addRow(("wavelength (angstrom)", f"{self.calc.wavelength:.5f}"))
        table.addRow(("calc engine", self.calc.engine.name))
        table.addRow(("mode", self.calc.engine.mode))

        pt = pyRestTable.Table()
        pt.labels = "name value".split()
        if self.calc._axis_name_to_original:
            pt.addLabel("original name")
        for item in self.real_positioners:
            row = [item.attr_name, f"{item.position:.5f}"]
            k = self.calc._axis_name_to_original.get(item.attr_name)
            if k is not None:
                row.append(k)
            pt.addRow(row)
        table.addRow(("positions", addTable(pt)))

        t = self.show_constraints(printing=False)
        table.addRow(("constraints", addTable(t)))

        if all_samples:
            samples = self.calc._samples.values()
        else:
            samples = [self.calc._sample]
        for sample in samples:
            t = pyRestTable.Table()
            t.labels = "term value".split()
            nm = sample.name
            if all_samples and sample == self.calc.sample:
                nm += " (*)"

            # fmt: off
            t.addRow(
                (
                    "unit cell edges",
                    Package(
                        **{
                            k: getattr(sample.lattice, k)
                            for k in "a b c".split()
                        }
                    ),
                )
            )
            t.addRow(
                (
                    "unit cell angles",
                    Package(
                        **{
                            k: getattr(sample.lattice, k)
                            for k in "alpha beta gamma".split()
                        }
                    ),
                )
            )
            # fmt: on

            for i, ref in enumerate(sample._sample.reflections_get()):
                h, k, l = ref.hkl_get()
                pos_arr = ref.geometry_get().axis_values_get(self.calc._units)
                t.addRow((f"ref {i+1} (hkl)", Package(**dict(h=h, k=k, l=l))))
                # fmt: off
                t.addRow(
                    (
                        f"ref {i+1} positioners",
                        Package(
                            **{
                                k: f"{v:.5f}"
                                for k, v in zip(
                                    self.calc.physical_axis_names, pos_arr
                                )
                            }
                        ),
                    )
                )
                # fmt: on

            t.addRow(("[U]", sample.U))
            t.addRow(("[UB]", sample.UB))

            table.addRow((f"sample: {nm}", addTable(t)))

        if printing:
            print(table)

        return table

    def wh(self, printing=True):
        """
        Report (brief) where is the diffractometer.

        EXAMPLE::

            In [1]: from hkl import SimulatedK4CV

            In [2]: k4cv = SimulatedK4CV('', name='k4cv')

            In [3]: k4cv.wh()
            ===================== ========= =========
            term                  value     axis_type
            ===================== ========= =========
            diffractometer        k4cv
            sample name           main
            energy (keV)          8.05092
            wavelength (angstrom) 1.54000
            calc engine           hkl
            mode                  bissector
            h                     0.0       pseudo
            k                     0.0       pseudo
            l                     0.0       pseudo
            komega                0         real
            kappa                 0         real
            kphi                  0         real
            tth                   0         real
            ===================== ========= =========

            Out[3]: <pyRestTable.rest_table.Table at 0x7f55c4775cd0>

        compare with similar function in SPEC::

            1117.KAPPA> wh
            H K L =  0  0  1.7345
            Alpha = 20  Beta = 20  Azimuth = 90
            Omega = 32.952  Lambda = 1.54
            Two Theta       Theta         Chi         Phi     K_Theta       Kappa       K_Phi
            40.000000   20.000000   90.000000   57.048500   77.044988  134.755995  114.093455

        """
        table = pyRestTable.Table()
        table.labels = "term value axis_type".split()
        table.addRow(("diffractometer", self.name, ""))
        table.addRow(("sample name", self.calc.sample.name, ""))
        table.addRow(("energy (keV)", f"{self.calc.energy:.5f}", ""))
        table.addRow(("wavelength (angstrom)", f"{self.calc.wavelength:.5f}", ""))
        table.addRow(("calc engine", self.calc.engine.name, ""))
        table.addRow(("mode", self.calc.engine.mode, ""))

        pseudo_axes = [v.attr_name for v in self._pseudo]
        real_axes = [v.attr_name for v in self._real]
        for k in self._sig_attrs.keys():
            v = getattr(self, k)
            if not issubclass(v.__class__, PositionerBase):
                continue
            if k in real_axes:
                label = "real"
            elif k in pseudo_axes:
                label = "pseudo"
            else:
                label = "additional"
            table.addRow((k, v.position, label))

        if printing:
            print(table)

        return table



"""
Calculation support for diffractometers

.. autosummary::

    ~default_decision_function
    ~UnreachableError
    ~CalcRecip

**Specific Geometries**

.. autosummary::

    ~CalcE4CH
    ~CalcE4CV
    ~CalcE6C
    ~CalcK4CV
    ~CalcK6C
    ~CalcPetra3_p09_eh2
    ~CalcSoleilMars
    ~CalcSoleilSiriusKappa
    ~CalcSoleilSiriusTurret
    ~CalcSoleilSixsMed1p2
    ~CalcSoleilSixsMed2p2
    ~CalcSoleilSixsMed2p3
    ~CalcZaxis

"""

import logging
import functools
from collections import OrderedDict
from threading import RLock

import numpy as np

from .engine import Engine, CalcParameter
from .sample import HklSample
from . import util
from .util import libhkl
from .context import UsingEngine

__all__ = """
    A_KEV
    CalcE4CH
    CalcE4CV
    CalcE6C
    CalcK4CV
    CalcK6C
    CalcPetra3_p09_eh2
    CalcRecip
    CalcSoleilMars
    CalcSoleilSiriusKappa
    CalcSoleilSiriusTurret
    CalcSoleilSixsMed1p2
    CalcSoleilSixsMed2p2
    CalcSoleilSixsMed2p3
    CalcZaxis
    default_decision_function
    NM_KEV
    UnreachableError
""".split()
logger = logging.getLogger(__name__)

# per NIST publication of CODATA Fundamental Physical Constants
# https://www.nist.gov/programs-projects/codata-values-fundamental-physical-constants
A_KEV = 12.3984198  # 1 Angstrom ~= 12.39842 keV
# A_KEV = 12.39842  # 1 Angstrom ~= 12.39842 keV
NM_KEV = A_KEV * 0.1  # 1 nm ~= 1.239842 keV
# Note: NM_KEV is not used by hklpy now, remains here as legacy


def default_decision_function(position, solutions):
    """The default decision function - returns the first solution."""
    return solutions[0]


# This is used below by CalcRecip.
def _locked(func):
    """a decorator for running a method with the instance's lock"""

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        with self._lock:
            return func(self, *args, **kwargs)

    return wrapped


# This is used below by CalcRecip.
def _keep_physical_position(func):
    """
    decorator: stores/restores the physical motor position during calculations
    """

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        with self._lock:
            initial_pos = self.physical_positions
            try:
                return func(self, *args, **kwargs)
            finally:
                self.physical_positions = initial_pos

    return wrapped


class UnreachableError(ValueError):
    """Position is unreachable.

    Attributes
    ----------
    pseudo : sequence
        Last reachable pseudo position in the trajectory
    physical : sequence
        Corresponding physical motor positions
    """

    def __init__(self, msg, pseudo, physical):
        super().__init__(msg)
        self.pseudo = pseudo
        self.physical = physical


class CalcRecip(object):
    """Reciprocal space calculations

    .. autosummary::

        ~add_sample
        ~calc_linear_path
        ~energy
        ~engine
        ~engine_locked
        ~engines
        ~forward
        ~forward_iter
        ~geometry_name
        ~get_path
        ~inverse
        ~inverted_axes
        ~new_sample
        ~parameters
        ~physical_axes
        ~physical_axis_names
        ~physical_positions
        ~Position
        ~pseudo_axes
        ~pseudo_axis_names
        ~pseudo_positions
        ~sample
        ~sample_name
        ~units
        ~update
        ~wavelength

    Parameters
    ----------
    dtype : str
        Diffractometer type (usually specified by a subclass)
    engine : str, optional
        'hkl', for example
    sample : str, optional
        Default sample name (default: 'main')
    lattice : Lattice, optional
        Lattice to use with the default sample
    degrees : bool, optional
        Use degrees instead of radians (default: True)
    units : {'user', }
        The type of units to use internally
    lock_engine : bool, optional
        Don't allow the engine to be changed during
        the life of this object
    inverted_axes : list, optional
        Names of axes to invert the sign
    """

    def __init__(
        self,
        dtype,
        engine="hkl",
        sample="main",
        lattice=None,
        degrees=True,
        units="user",
        lock_engine=False,
        inverted_axes=None,
    ):

        self._engine = None  # set below with property
        self._detector = util.new_detector()
        self._degrees = bool(degrees)
        self._sample = None
        self._samples = {}
        self._unit_name = units
        self._units = util.units[self._unit_name]
        self._lock_engine = bool(lock_engine)
        self._lock = RLock()
        self._axis_name_to_renamed = {}
        self._axis_name_to_original = {}
        self._inverted_axes = inverted_axes or []

        try:
            self._factory = libhkl.factories()[dtype]
        except KeyError:
            types = ", ".join(util.diffractometer_types)
            raise ValueError(f"Invalid diffractometer type {types!r};  choose from: {types}")

        self._geometry = self._factory.create_new_geometry()
        self._engine_list = self._factory.create_new_engine_list()

        if sample is not None:
            if isinstance(sample, HklSample):
                if lattice is not None:
                    sample.lattice = lattice
                self.add_sample(sample)
            else:
                self.new_sample(sample, lattice=lattice)

        self.engine = engine

    @property
    def Position(self):
        """Dynamically-generated physical motor position class"""
        # I know, I know, could be done more cleanly...
        name = f"Pos{self.__class__.__name__}"
        return util.get_position_tuple(self.physical_axis_names, class_name=name)

    @property
    def wavelength(self):
        """The wavelength associated with the geometry, in angstrom"""
        return self._geometry.wavelength_get(self._units)

    @wavelength.setter
    def wavelength(self, wavelength):
        self._geometry.wavelength_set(wavelength, self._units)

    @property
    def energy(self):
        """The energy associated with the geometry, in keV"""
        return A_KEV / self.wavelength

    @energy.setter
    def energy(self, energy):
        self.wavelength = A_KEV / energy

    @property
    def engine_locked(self):
        """If set, do not allow the engine to be changed post-initialization"""
        return self._lock_engine

    @property
    def engine(self):
        """ """
        return self._engine

    @engine.setter
    @_locked
    def engine(self, engine):
        if engine is self._engine:
            return

        if self._lock_engine and self._engine is not None:
            raise ValueError("Engine is locked on this %s instance" % self.__class__.__name__)

        if isinstance(engine, libhkl.Engine):
            self._engine = engine
        else:
            engines = self.engines
            try:
                self._engine = engines[engine]
            except KeyError:
                raise ValueError("Unknown engine name or type")

        self._re_init()

    @property
    def geometry_name(self):
        """Name of this geometry, as defined in **libhkl**."""
        return self._geometry.name_get()

    def _get_sample(self, name):
        if isinstance(name, libhkl.Sample):
            return name

        return self._samples[name]

    @property
    def sample_name(self):
        """The name of the currently selected sample."""
        return self._sample.name

    @sample_name.setter
    @_locked
    def sample_name(self, new_name):
        sample = self._sample
        sample.name = new_name

    @property
    def sample(self):
        """Currently selected sample."""
        return self._sample

    @sample.setter
    @_locked
    def sample(self, sample):
        if sample is self._sample:
            return
        elif sample == self._sample.name:
            return

        if isinstance(sample, HklSample):
            if sample not in self._samples.values():
                self.add_sample(sample, select=False)
        elif sample in self._samples:
            name = sample
            sample = self._samples[name]
        else:
            raise ValueError("Unknown sample type (expected HklSample)")

        self._sample = sample
        self._re_init()

    def add_sample(self, sample, select=True):
        """Add an HklSample

        Parameters
        ----------
        sample : HklSample instance
            The sample name, or optionally an already-created HklSample
            instance
        select : bool, optional
            Select the sample to focus calculations on
        """
        if not isinstance(sample, (HklSample, libhkl.Sample)):
            raise ValueError("Expected either an HklSample or a Sample instance")

        if isinstance(sample, libhkl.Sample):
            sample = HklSample(calc=self, sample=sample, units=self._unit_name)

        if sample.name in self._samples:
            raise ValueError('Sample of name "%s" already exists' % sample.name)

        self._samples[sample.name] = sample
        if select:
            self._sample = sample
            self._re_init()

        return sample

    def new_sample(self, name, select=True, **kwargs):
        """Convenience function to add a sample by name

        Keyword arguments are passed to the new HklSample initializer.

        Parameters
        ----------
        name : str
            The sample name
        select : bool, optional
            Select the sample to focus calculations on
        """
        units = kwargs.pop("units", self._unit_name)
        sample = HklSample(self, sample=libhkl.Sample.new(name), units=units, **kwargs)

        return self.add_sample(sample, select=select)

    @_locked
    def _re_init(self):
        if self._engine is None:
            return

        if self._geometry is None or self._detector is None or self._sample is None:
            raise ValueError("Not all parameters set (geometry, detector, sample)")
            # pass
        else:
            self._engine_list.init(self._geometry, self._detector, self._sample.hkl_sample)

    @property
    def engines(self):
        return dict(
            (engine.name_get(), Engine(self, engine, self._engine_list))
            for engine in self._engine_list.engines_get()
        )

    @property
    def parameters(self):
        return self._engine.parameters

    @property
    def physical_axis_names(self):
        if self._axis_name_to_renamed:
            return list(self._axis_name_to_renamed.values())
        else:
            return self._geometry.axis_names_get()

    @physical_axis_names.setter
    def physical_axis_names(self, axis_name_map):
        """Set a persistent re-map of physical axis names

        Resets `inverted_axes`.

        Parameters
        ----------
        axis_name_map : dict
            {orig_axis_1: new_name_1, ...}
        """
        internal_axis_names = self._geometry.axis_names_get()
        if set(axis_name_map.keys()) != set(internal_axis_names):
            raise ValueError("Every axis name has to have a remapped name")

        self._axis_name_to_original = OrderedDict((axis_name_map[axis], axis) for axis in internal_axis_names)
        self._axis_name_to_renamed = OrderedDict((axis, axis_name_map[axis]) for axis in internal_axis_names)

        self._inverted_axes = []

    @property
    def inverted_axes(self):
        """The physical axis names to invert"""
        return self._inverted_axes

    @inverted_axes.setter
    def inverted_axes(self, to_invert):
        for axis in to_invert:
            assert axis in self.physical_axis_names

        self._inverted_axes = to_invert

    def _invert_physical_positions(self, pos):
        """Invert the physical axis positions based on the settings

        Parameters
        ----------
        pos : OrderedDict
            NOTE: Modified in-place
        """
        for axis in self._inverted_axes:
            pos[axis] = -pos[axis]
        return pos

    @property
    def physical_positions(self):
        """Physical (real) motor positions"""
        pos = self.physical_axes
        if self._inverted_axes:
            pos = self._invert_physical_positions(pos)

        return self.Position(*pos.values())

    @physical_positions.setter
    @_locked
    def physical_positions(self, positions):
        if self._inverted_axes:
            pos = self.Position(*positions)._asdict()
            pos = self._invert_physical_positions(pos)
            positions = list(pos.values())

        # Set the physical motor positions and calculate the pseudo ones
        self._geometry.axis_values_set(positions, self._units)
        self.update()

    @property
    def physical_axes(self):
        """Physical (real) motor positions as an OrderedDict"""
        return OrderedDict(
            # fmt: off
            zip(
                self.physical_axis_names,
                self._geometry.axis_values_get(self._units),
            )
            # fmt: on
        )

    @property
    def pseudo_axis_names(self):
        """Pseudo axis names from the current engine"""
        return self._engine.pseudo_axis_names

    @property
    def pseudo_positions(self):
        """Pseudo axis positions/values from the current engine"""
        return self._engine.pseudo_positions

    @property
    def pseudo_axes(self):
        """Ordered dictionary of axis name to position"""
        return self._engine.pseudo_axes

    def update(self):
        """Calculate the pseudo axis positions from the real axis positions"""
        return self._engine.update()

    def _get_axis_by_name(self, name):
        """Given an axis name, return the HklParameter

        Parameters
        ----------
        name : str
            If a name map is specified, this is the mapped name.
        """
        name = self._axis_name_to_original.get(name, name)
        return self._geometry.axis_get(name)

    @property
    def units(self):
        """The units used for calculations"""
        return self._unit_name

    def __getitem__(self, axis):
        return CalcParameter(
            self._get_axis_by_name(axis),
            units=self._unit_name,
            name=axis,
            inverted=axis in self._inverted_axes,
            geometry=self._geometry,
        )

    def __setitem__(self, axis, value):
        param = self[axis]
        param.value = value

    @_keep_physical_position
    def forward_iter(self, start, end, max_iters, *, threshold=0.99, decision_fcn=None):
        """Iteratively attempt to go from a pseudo start -> end position

        For every solution failure, the position is moved back.
        For every success, the position is moved closer to the destination.

        After up to max_iters, the position can be reached, the solutions will
        be returned. Otherwise, ValueError will be raised stating the last
        position that was reachable and the corresponding motor positions.

        Parameters
        ----------
        start : Position
        end : Position
        max_iters : int
            Maximum number of iterations
        threshold : float, optional
            Normalized proximity to `end` position to stop iterating
        decision_fcn : callable, optional
            Function to choose a solution from several. Defaults to picking the
            first solution. The signature of the function should be as follows:

            >>  def decision(pseudo_position, solution_list):
            >>      return solution_list[0]

        Returns
        -------
        solutions : list

        Raises
        ------
        UnreachableError (ValueError)
            Position cannot be reached
            The last valid HKL position and motor positions are accessible
            in this exception instance.
        """
        start = np.array(start)
        end = np.array(end)

        if decision_fcn is None:
            decision_fcn = default_decision_function

        min_t = 0.0

        t = 1.0
        iters = 0
        valid_pseudo = None
        valid_real = None
        while iters < max_iters:
            try:
                pos = (1.0 - t) * start + t * end
                self.engine.pseudo_positions = pos
            except ValueError:
                # couldn't calculate - step back half-way
                t = (min_t + t) / 2.0
            else:
                if t > min_t:
                    min_t = t

                # successful calculation - move forward
                t = (t + 1) / 2.0

                valid_pseudo = self.engine.pseudo_positions
                valid_real = decision_fcn(valid_pseudo, self.engine.solutions)
                self.physical_positions = valid_real

                if t >= threshold:
                    break

            iters += 1

        try:
            self.engine.pseudo_positions = end
            return self.engine.solutions
        except ValueError:
            raise UnreachableError(
                f"Unable to solve. iterations={iters}/{max_iters}\n"
                f"Last valid position: {valid_pseudo}\n{valid_real} ",
                pseudo=valid_pseudo,
                physical=valid_real,
            )

    @_keep_physical_position
    def forward(self, position, engine=None):
        """Calculate real positions from pseudo positions."""

        with UsingEngine(self, engine):
            if self.engine is None:
                raise ValueError("Engine unset")

            self.engine.pseudo_positions = position
            return self.engine.solutions

    @_keep_physical_position
    def inverse(self, real):
        """Calculate pseudo positions from real positions."""
        self.physical_positions = real
        # self.update()  # done implicitly in setter
        return self.pseudo_positions

    def calc_linear_path(self, start, end, n, num_params=0, **kwargs):
        """
        Construct a trajectory from start to end with n steps (n+1 points).

        Example (for ``hkl`` engine)::

            >>> e4cv.calc.calc_linear_path((1,1,0), (1,-1,0), 4, num_params=3)
            [(1.0, 1.0, 0.0),
            (1.0, 0.5, 0.0),
            (1.0, 0.0, 0.0),
            (1.0, -0.5, 0.0),
            (1.0, -1.0, 0.0)]
        """

        # from start to end, in a linear path
        singles = [np.linspace(start[i], end[i], n + 1) for i in range(num_params)]

        return list(zip(*singles))

    def _get_path_fcn(self, path_type):
        try:
            return getattr(self, "calc_%s_path" % (path_type))
        except AttributeError:
            raise ValueError("Invalid path type specified (%s)" % path_type)

    def get_path(self, start, end=None, n=100, path_type="linear", **kwargs):
        """ """
        num_params = len(self.pseudo_axis_names)

        start = np.array(start)

        path_fcn = self._get_path_fcn(path_type)

        if end is not None:
            end = np.array(end)
            if start.size == end.size == num_params:
                return path_fcn(start, end, n, num_params=num_params, **kwargs)

        else:
            positions = np.array(start)
            if positions.ndim == 1 and positions.size == num_params:
                # single position
                return [list(positions)]
            elif positions.ndim == 2:
                if positions.shape[0] == 1 and positions.size == num_params:
                    # [[h, k, l], ]
                    return [positions[0]]
                elif positions.shape[0] == num_params:
                    # [[h, k, l], [h, k, l], ...]
                    return [positions[i, :] for i in range(num_params)]

        raise ValueError("Invalid set of %s positions" % ", ".join(self.pseudo_axis_names))

    def __call__(
        # fmt: off
        self,
        start,
        end=None,
        n=100,
        engine=None,
        path_type="linear",
        **kwargs,
        # fmt: on
    ):

        with UsingEngine(self, engine):
            for pos in self.get_path(start, end=end, n=n, path_type=path_type, **kwargs):
                yield self.forward(pos, engine=None, **kwargs)

    def _repr_info(self):
        r = [
            f"engine={self.engine.name!r}",
            f"detector={self._detector!r}",
            f"sample={self._sample!r}",
            f"samples={self._samples!r}",
        ]

        return r

    def __repr__(self):
        return f"{self.__class__.__name__} ({', '.join(self._repr_info())})"

    def __str__(self):
        return repr(self)


class CalcE4CH(CalcRecip):
    """Geometry: E4CH"""

    def __init__(self, **kwargs):
        super().__init__("E4CH", **kwargs)


class CalcE4CV(CalcRecip):
    """Geometry: E4CV"""

    def __init__(self, **kwargs):
        super().__init__("E4CV", **kwargs)


class CalcE6C(CalcRecip):
    """Geometry: E6C"""

    def __init__(self, **kwargs):
        super().__init__("E6C", **kwargs)


class CalcK4CV(CalcRecip):
    """Geometry: K4CV"""

    def __init__(self, **kwargs):
        super().__init__("K4CV", **kwargs)


class CalcK6C(CalcRecip):
    """Geometry: K6C"""

    def __init__(self, **kwargs):
        super().__init__("K6C", **kwargs)


class CalcPetra3_p09_eh2(CalcRecip):
    """Geometry: PETRA3 P09 EH2"""

    def __init__(self, **kwargs):
        super().__init__("PETRA3 P09 EH2", **kwargs)


class CalcSoleilMars(CalcRecip):
    """Geometry: SOLEIL MARS"""

    def __init__(self, **kwargs):
        super().__init__("SOLEIL MARS", **kwargs)


class CalcSoleilSiriusKappa(CalcRecip):
    """Geometry: SOLEIL SIRIUS KAPPA"""

    def __init__(self, **kwargs):
        super().__init__("SOLEIL SIRIUS KAPPA", **kwargs)


class CalcSoleilSiriusTurret(CalcRecip):
    """Geometry: SOLEIL SIRIUS TURRET"""

    def __init__(self, **kwargs):
        super().__init__("SOLEIL SIRIUS TURRET", **kwargs)


class CalcSoleilSixsMed1p2(CalcRecip):
    """Geometry: SOLEIL SIXS MED1+2"""

    def __init__(self, **kwargs):
        super().__init__("SOLEIL SIXS MED1+2", **kwargs)


class CalcSoleilSixsMed2p2(CalcRecip):
    """Geometry: SOLEIL SIXS MED2+2"""

    def __init__(self, **kwargs):
        super().__init__("SOLEIL SIXS MED2+2", **kwargs)


class CalcSoleilSixsMed2p3(CalcRecip):
    def __init__(self, **kwargs):
        super().__init__("SOLEIL SIXS MED2+3", **kwargs)


class CalcZaxis(CalcRecip):
    """Geometry: ZAXIS"""

    def __init__(self, **kwargs):
        super().__init__("ZAXIS", **kwargs)



"""
Utility functions and structures.

.. autosummary::

    ~Constraint
    ~diffractometer_types
    ~get_package_info
    ~get_position_tuple
    ~Lattice
    ~list_orientation_runs
    ~new_detector
    ~restore_constraints
    ~restore_energy
    ~restore_orientation
    ~restore_reflections
    ~restore_sample
    ~restore_UB
    ~run_orientation_info
    ~software_versions

Also provides `SI_LATTICE_PARAMETER` as defined by the
*2018 CODATA recommended lattice parameter of silicon*. [#]_

.. [#] https://physics.nist.gov/cgi-bin/cuu/Value?asil
"""

from __future__ import print_function
from collections import defaultdict, namedtuple
import gi
import logging
import numpy as np
import pandas as pd
import subprocess

gi.require_version("Hkl", "5.0")
from gi.repository import Hkl as libhkl
from gi.repository import GLib  # noqa: F401

__all__ = """
    Constraint
    diffractometer_types
    get_package_info
    get_position_tuple
    Lattice
    list_orientation_runs
    new_detector
    restore_constraints
    restore_energy
    restore_orientation
    restore_reflections
    restore_sample
    restore_UB
    run_orientation_info
    SI_LATTICE_PARAMETER
    software_versions
""".split()
logger = logging.getLogger(__name__)


# when getting software package versions
DEFAULT_PACKAGE_LIST = "hkl hklpy gobject-introspection".split()

# 2018 CODATA recommended lattice parameter of silicon, Angstrom.
# see: https://physics.nist.gov/cgi-bin/cuu/Value?asil
SI_LATTICE_PARAMETER = 5.431020511
# also provide the reported uncertainty, in case anyone is interested
SI_LATTICE_PARAMETER_UNCERTAINTY = 0.000000089


def new_detector(dtype=0):
    """Create a new HKL-library detector"""
    return libhkl.Detector.factory_new(libhkl.DetectorType(dtype))


if libhkl:
    diffractometer_types = tuple(sorted(libhkl.factories().keys()))
    UserUnits = libhkl.UnitEnum.USER
    DefaultUnits = libhkl.UnitEnum.DEFAULT

    # fmt: off
    units = {
        "user": UserUnits,
        "default": DefaultUnits
    }
    # fmt: on
else:
    diffractometer_types = ()
    units = {}


def to_numpy(mat):
    """Convert an hkl ``Matrix`` to a numpy ndarray

    Parameters
    ----------
    mat : Hkl.Matrix

    Returns
    -------
    ndarray
    """
    if isinstance(mat, np.ndarray):
        return mat

    ret = np.zeros((3, 3))
    for i in range(3):
        for j in range(3):
            ret[i, j] = mat.get(i, j)

    return ret


def to_hkl(arr):
    """Convert a numpy ndarray to an hkl ``Matrix``

    Parameters
    ----------
    arr : ndarray

    Returns
    -------
    Hkl.Matrix
    """
    if isinstance(arr, libhkl.Matrix):
        return arr

    arr = np.array(arr)

    hklm = hkl_euler_matrix(0, 0, 0)
    hklm.init(*arr.flatten())
    return hklm


def hkl_euler_matrix(euler_x, euler_y, euler_z):
    """Convert into matrix form."""
    return libhkl.Matrix.new_euler(euler_x, euler_y, euler_z)


def _gi_info(gi_val):
    def get(attr):
        try:
            getter = getattr(gi_val, attr)
            # inspect.signature doesn't work on gi functions...
            return getter()
        except Exception as ex:
            try:
                return getter(units["user"])
            except Exception:
                return f"({ex.__class__.__name__}: {ex})"

    return {attr: get(attr) for attr in dir(gi_val) if attr.endswith("_get")}


class Constraint:
    """
    Limitations on acceptable positions from computed forward() solutions.

    Parameters
    ----------
    low_limit : float
        Minimum acceptable solution for position.
    high_limit : float
        Maximum acceptable solution for position.
    value : float
        Constant value when ``fit=False``.
    fit : bool
        ``True`` when axis will be fitted, otherwise, hold position to ``value``.


    note: Patterned on collections.namedtuple
    """

    def __init__(self, low_limit, high_limit, value, fit):
        self.low_limit = float(low_limit)
        self.high_limit = float(high_limit)
        self.value = float(value)
        self.fit = bool(fit)

        self._fields = "low_limit high_limit value fit".split()
        # fmt: off
        _fields = ", ".join(
            name + "=" + repr(getattr(self, name))
            for name in self._fields
        )
        # fmt: on
        self._repr_fmt = f"({_fields})"

    def _asdict(self):
        "Return a new dict which maps field names to their values."
        return dict(zip(self._fields, self))

    class _ConstraintIterator:
        "Iterator"

        def __init__(self, constraint):
            self.constraint = constraint
            self._index = 0

        def __next__(self):
            if self._index < len(self.constraint._fields):
                c = getattr(self.constraint, self.constraint._fields[self._index])
                self._index += 1
                return c
            else:
                raise StopIteration

    def __iter__(self):
        "Iterate through the fields."
        return self._ConstraintIterator(self)

    def __repr__(self):
        "Return a nicely formatted representation string."
        content = "(" + ", ".join([f"{k}={getattr(self, k)}" for k in self._fields]) + ")"
        return self.__class__.__name__ + content


Lattice = namedtuple("LatticeTuple", "a b c alpha beta gamma")


_position_tuples = {}


def get_position_tuple(axis_names, class_name="Position"):
    """Return a namedtuple with the positions."""
    global _position_tuples

    key = frozenset(axis_names)
    if key not in _position_tuples:
        _position_tuples[key] = namedtuple(class_name, tuple(axis_names))

    return _position_tuples[key]


def list_orientation_runs(catalog, *args, limit=20):
    """
    List the runs with orientation information.

    Returns
    -------
    orientation information: Pandas DataFrame object

    Parameters
    ----------
    *args : [str]
        A list of additional data column names to be displayed,
        corresponding to the names of available orientation
        information in the descriptor document.

        Example::

            list_orientation_runs("class_name", energy", "energy_units", "lattice")

    catalog : object
        Instance of a databroker catalog.
    limit : int
        Limit the list to the first ``limit`` rows. (default=20)
    """
    buffer = []
    _count = 0
    default_columns = "sample_name diffractometer_name geometry_name".split()
    display_columns = default_columns + list(args)
    for run in catalog.v2.values():
        info = run_orientation_info(run)
        if len(info):
            scan_id = run.metadata["start"]["scan_id"]
            uid = run.name[:7]
            for device in sorted(info.keys()):
                orientation = info[device]
                row = dict(scan_id=scan_id)
                for f in display_columns:
                    if f in orientation:
                        row[f] = orientation[f]
                row["uid"] = uid
                buffer.append(row)
                _count += 1
                if _count >= limit:
                    break
        if _count >= limit:
            break
    return pd.DataFrame(buffer)


def run_orientation_info(run):
    """
    Return a dictionary with orientation information in this run.

    Dictionary is keyed by "detector" name (in case more than one
    diffractometer was added as a "detector" to the run).

    The orientation information is found in the descriptor document
    of the primary stream.

    Parameters
    ----------
    run : from Databroker
        A Bluesky run, from databroker v2, such as ``cat.v2[-1]``.
    """
    devices = {}
    run_conf = run.primary.config
    for device in sorted(run_conf):
        conf = run_conf[device].read()
        if f"{device}_orientation_attrs" in conf:
            # fmt:off
            devices[device] = {
                item[len(device)+1:]: conf[item].to_dict()["data"][0]
                for item in conf
            }
            # fmt:on
    return devices


def _smart_signal_update(value, signal):
    """Write value to signal if not equal.  Not a plan."""
    if signal.get() != value:
        signal.put(value)


def _check_geometry(orientation, diffractometer):
    """
    Check that geometry of recovered orientation matches diffractometer.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.

    Raise ValueError if mismatched.
    """
    if orientation["geometry_name"] != diffractometer.geometry_name.get():
        raise ValueError(
            "Geometries do not match:"
            f" Orientation={orientation['geometry_name']},"
            f" {diffractometer.name}={diffractometer.geometry_name.get()},"
            " will not restore."
        )


def restore_constraints(orientation, diffractometer):
    """
    Restore any constraints from orientation information.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    _check_geometry(orientation, diffractometer)

    # fmt:off
    if len(orientation["_constraints"]):
        con_dict = {
            k: Constraint(*v)
            for k, v in zip(orientation["_reals"], orientation["_constraints"])
        }
        diffractometer.apply_constraints(con_dict)
    # fmt:on


def restore_energy(orientation, diffractometer):
    """
    Restore energy from orientation information.

    NOTE: This makes blocking calls so do not use in bluesky plan.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    for attr in "energy energy_units energy_offset".split():
        _smart_signal_update(orientation[attr], getattr(diffractometer, attr))


def restore_reflections(orientation, diffractometer):
    """
    Restore any reflections from orientation information.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    _check_geometry(orientation, diffractometer)
    # remember this wavelength
    wavelength0 = diffractometer.calc.wavelength

    # short aliases
    pseudos = orientation["_pseudos"]
    reals = orientation["_reals"]
    orientation_reflections = []

    for ref_base in orientation["reflections_details"]:
        # every reflection has its own wavelength
        diffractometer.calc.wavelength = ref_base["wavelength"]

        # Order of the items is important.
        # Can't just use the dictionaries in ``orientation``.
        # Get the canonical order from the orientation data.
        miller_indices = [ref_base["reflection"][key] for key in pseudos]
        positions = [ref_base["position"][key] for key in reals]
        ppp = namedtuple("PositionTuple", tuple(reals))(*positions)

        # assemble the final form of the reflection for add_reflection()
        reflection = tuple([*miller_indices, ppp])
        r = diffractometer.calc.sample.add_reflection(*reflection)
        if ref_base["orientation_reflection"]:
            orientation_reflections.append(r)

    if len(orientation_reflections) > 1:
        # compute **UB** from the last two orientation reflections
        diffractometer.calc.sample.compute_UB(*orientation_reflections[-2:])

    # restore previous wavelength
    if diffractometer.calc.wavelength != wavelength0:
        diffractometer.calc.wavelength = wavelength0


def restore_orientation(orientation, diffractometer):
    """
    Restore all orientation information.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    _check_geometry(orientation, diffractometer)
    restore_energy(orientation, diffractometer)
    restore_sample(orientation, diffractometer)
    restore_constraints(orientation, diffractometer)
    restore_reflections(orientation, diffractometer)


def restore_sample(orientation, diffractometer):
    """
    Restore sample & lattice from orientation information.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    nm = orientation["sample_name"]
    lattice = orientation["lattice"]
    if nm not in diffractometer.calc._samples:
        diffractometer.calc.new_sample(nm, lattice=lattice)
    else:
        raise ValueError(f"Sample '{nm}' already exists in {diffractometer.name}.")


def restore_UB(orientation, diffractometer):
    """
    Restore **UB** matrix from orientation information.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    _check_geometry(orientation, diffractometer)
    _smart_signal_update(orientation["UB"], diffractometer.UB)


def _installed_package_information():
    """Index information about packages known by conda and/or pip."""
    packages = defaultdict(dict)

    def run(tool):
        try:
            # Python 3.7+
            s = subprocess.run([tool, "list"], capture_output=True)
        except TypeError:
            # Python < 3.7
            s = subprocess.run([tool, "list"], stdout=subprocess.PIPE)
        return s

    for tool in "conda pip".split():
        s = run(tool)
        for line in s.stdout.splitlines():
            if not line.decode().startswith("#"):
                args = line.decode().strip().split()
                key = args[0]
                packages[key]["version"] = args[1]
                packages[key][tool] = True
                if tool == "conda":
                    packages[key]["build"] = args[2]
                    packages[key]["conda"] = True
                    if len(args) == 4:
                        packages[key]["channel"] = args[3]

                elif tool == "pip":
                    if len(args) == 3:
                        packages[key]["location"] = args[2]

    return packages


# cache of discovered installed package information
_package_info = None


def get_package_info(package_name):
    """Return dict of information about installed package, by name."""
    global _package_info
    if _package_info is None:
        # index the known packages
        # This is not expected to change once the process has started.
        _package_info = _installed_package_information()
    return _package_info.get(package_name)


def software_versions(keys=[]):
    """
    Report the package versions, in a dictionary.

    EXAMPLE::

        In [1]: import gi
        ...:
        ...: gi.require_version("Hkl", "5.0")
        ...:
        ...: import hkl.util

        In [2]: hkl.util.software_versions()
        Out[2]:
        {'hkl': '5.0.0.2173',
        'hklpy': '0.3.16+131.ga5a449a.dirty',
        'gobject-introspection': '1.68.0'}

    Here, it shows (albeit indirectly) *Hkl 5.0.0 tag 2173*.  Proceed to the Hkl
    source repository, list of tags [#]_, and find tag ``2173`` [#]_.

    .. [#] Hkl source tags: https://repo.or.cz/hkl.git/refs
    .. [#] Hkl tag 2173: https://repo.or.cz/hkl.git/tree/refs/tags/v5.0.0.2173)
    """
    if keys is None or len(keys) == 0:
        keys = DEFAULT_PACKAGE_LIST
    v_dict = {}
    for key in keys:
        info = get_package_info(key)
        if info is not None:
            v_dict[key] = info.get("version")
    return v_dict



"""
:mod:`hkl` - HKL diffractometer support
=======================================

.. module:: hkl
   :synopsis:

"""

import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

from ._version import get_versions  # noqa: F402, E402

__version__ = get_versions()["version"]
del get_versions

# import shortcuts

from .calc import A_KEV, UnreachableError  # noqa: F401, F402, E402

from .geometries import (  # noqa: F401, F402, E402
    E4CH,
    E4CV,
    E6C,
    K4CV,
    K6C,
    Petra3_p09_eh2,
    SimMixin,
    SimulatedE4CV,
    SimulatedE6C,
    SimulatedK4CV,
    SimulatedK6C,
    SoleilMars,
    SoleilSiriusKappa,
    SoleilSiriusTurret,
    SoleilSixsMed1p2,
    SoleilSixsMed2p2,
    SoleilSixsMed2p3,
    Zaxis,
)

from .user import (  # noqa: F401, F402, E402
    cahkl,
    cahkl_table,
    calc_UB,
    change_sample,
    list_samples,
    new_sample,
    or_swap,
    pa,
    select_diffractometer,
    set_energy,
    setor,
    show_sample,
    show_selected_diffractometer,
    update_sample,
    wh,
)

from .util import (  # noqa: F401, F402, E402
    Constraint,
    diffractometer_types,
    get_position_tuple,
    Lattice,
    list_orientation_runs,
    new_detector,
    restore_constraints,
    restore_energy,
    restore_orientation,
    restore_reflections,
    restore_sample,
    restore_UB,
    run_orientation_info,
    SI_LATTICE_PARAMETER,
    software_versions,
)



#!/usr/bin/env python
from setuptools import setup, find_packages
import versioneer


setup(
    name="hklpy",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Controls for using diffractometers within the Bluesky Framework: https://blueskyproject.io",
    maintainer="prjemian",
    maintainer_email="prjemian+hklpy@gmail.com",
    url="https://github.com/bluesky/hklpy",
    license="BSD",
    packages=find_packages(),
)



import gi
import pytest


gi.require_version("Hkl", "5.0")
# NOTE: MUST call gi.require_version() BEFORE any import from hkl

import hkl
import hkl.calc
import hkl.diffract
import hkl.geometries
import hkl.user
import hkl.util


@pytest.mark.parametrize(
    "module_name, attribute",
    [
        ("calc", "A_KEV"),
        ("calc", "UnreachableError"),
        ("geometries", "E4CH"),
        ("geometries", "E4CV"),
        ("geometries", "E6C"),
        ("geometries", "K4CV"),
        ("geometries", "K6C"),
        ("geometries", "Petra3_p09_eh2"),
        ("geometries", "SimMixin"),
        ("geometries", "SimulatedE4CV"),
        ("geometries", "SimulatedE6C"),
        ("geometries", "SimulatedK4CV"),
        ("geometries", "SimulatedK6C"),
        ("geometries", "SoleilMars"),
        ("geometries", "SoleilSiriusKappa"),
        ("geometries", "SoleilSiriusTurret"),
        ("geometries", "SoleilSixsMed1p2"),
        ("geometries", "SoleilSixsMed2p2"),
        ("geometries", "SoleilSixsMed2p3"),
        ("geometries", "Zaxis"),
        ("user", "cahkl_table"),
        ("user", "cahkl"),
        ("user", "calc_UB"),
        ("user", "change_sample"),
        ("user", "list_samples"),
        ("user", "new_sample"),
        ("user", "or_swap"),
        ("user", "pa"),
        ("user", "select_diffractometer"),
        ("user", "set_energy"),
        ("user", "setor"),
        ("user", "show_sample"),
        ("user", "show_selected_diffractometer"),
        ("user", "update_sample"),
        ("user", "wh"),
        ("util", "Constraint"),
        ("util", "diffractometer_types"),
        ("util", "get_position_tuple"),
        ("util", "Lattice"),
        ("util", "list_orientation_runs"),
        ("util", "list_orientation_runs"),
        ("util", "restore_constraints"),
        ("util", "restore_energy"),
        ("util", "restore_orientation"),
        ("util", "restore_reflections"),
        ("util", "restore_reflections"),
        ("util", "restore_UB"),
        ("util", "run_orientation_info"),
        ("util", "software_versions"),
    ],
)
def test_shortcuts(module_name, attribute):
    assert hasattr(hkl, module_name)
    module = getattr(hkl, module_name)

    assert hasattr(module, attribute)
    actual = getattr(module, attribute)

    assert hasattr(hkl, attribute)
    shortcut = getattr(hkl, attribute)

    assert shortcut is actual



import gi

gi.require_version("Hkl", "5.0")
import pytest

import hkl.util
from packaging import version


NO_SUCH_PACKAGE_NAME = "no-such-package"


def test__package_info_states():
    assert hkl.util._package_info is None
    hkl.util.get_package_info("hkl")
    assert hkl.util._package_info is not None


@pytest.mark.parametrize(
    "package_name, minimum_version",
    [
        ("bluesky", "1.6"),
        ("gobject-introspection", "1.68.0"),
        ("hkl", "5.0.0"),
        ("hklpy", "0"),  # minimum test for unversioned use
        (NO_SUCH_PACKAGE_NAME, "---"),
        ("ophyd", "1.6"),
    ],
)
def test_get_package_info(package_name, minimum_version):
    v = hkl.util.get_package_info(package_name)
    if v is None:
        assert package_name == NO_SUCH_PACKAGE_NAME
    else:
        assert "version" in v
        v_string = v.get("version", "unknown")

        if package_name == "hkl":
            assert v_string.startswith("5.0.0.")

        v_package = version.parse(v_string)
        assert v_package >= version.parse(minimum_version)


@pytest.mark.parametrize(
    # fmt: off
    "case",
    [
        (None),
        ([]),
        ("")
    ],
    # fmt: on
)
def test_software_versions_default_list(case):
    v = hkl.util.software_versions(case)
    assert isinstance(v, dict)
    expected = sorted(hkl.util.DEFAULT_PACKAGE_LIST)
    assert sorted(v.keys()) == expected


@pytest.mark.parametrize(
    "package_name, minimum_version",
    [
        ("bluesky", "1.6"),
        ("gobject-introspection", "1.68.0"),
        ("hkl", "5.0.0"),
        ("hklpy", "0"),  # minimum test for unversioned use
        (NO_SUCH_PACKAGE_NAME, "---"),
        ("ophyd", "1.6"),
    ],
)
def test_software_versions_items(package_name, minimum_version):
    v = hkl.util.software_versions([package_name])
    if package_name in v:
        v_string = v[package_name]
        v_package = version.parse(v_string)
        assert v_package >= version.parse(minimum_version)
    else:
        assert package_name == NO_SUCH_PACKAGE_NAME
